from math import pi, sqrt

import numpy as np

from ..helpers import QuadratureScheme, backend_to_function, expand_symmetries

schemes = {}


def register(in_schemes):
    for scheme in in_schemes:
        schemes[scheme.__name__] = scheme


class E3r2Scheme(QuadratureScheme):
    def __init__(self, name, symmetry_data, degree, source, tol=1.0e-14):
        self.symmetry_data = symmetry_data
        points, weights = expand_symmetries(symmetry_data, dim=3)
        super().__init__(name, weights, points, degree, source, tol)
        self.domain = "E3r2"

    def integrate(self, f, dot=np.dot):
        flt = np.vectorize(float)
        ref_vol = sqrt(pi) ** 3
        return ref_vol * dot(f(flt(self.points)), flt(self.weights))

    def show(self, backend="vtk"):
        """Displays scheme for E_3^{r^2} quadrature."""
        backend_to_function[backend](
            self.points, self.weights, volume=8 * np.pi, edges=[]
        )


def get_good_scheme(degree):
    if degree <= 7:
        return {
            0: schemes["stroud_secrest_08a"](),
            1: schemes["stroud_secrest_08a"](),
            2: schemes["stroud_secrest_08a"](),
            3: schemes["stroud_secrest_08a"](),
            4: schemes["stroud_secrest_08a"](),
            5: schemes["stroud_secrest_08a"](),
            6: schemes["stroud_secrest_10b"](),
            7: schemes["stroud_secrest_10b"](),
        }[degree]
    return None



import pathlib

from sympy import Rational as frac

from ...helpers import article
from .._helpers import T2Scheme, _read, register

source = article(
    authors=["D.M. Williams", "L. Shunn", "A. Jameson"],
    title="Symmetric quadrature rules for simplexes based on sphere close packed lattice arrangements",
    journal="Journal of Computational and Applied Mathematics",
    volume="266",
    year="2014",
    pages="18–38",
    url="https://doi.org/10.1016/j.cam.2014.01.007",
)

this_dir = pathlib.Path(__file__).resolve().parent


def williams_shunn_jameson_1():
    d = {"centroid": [[1]]}
    return T2Scheme("Williams-Shunn-Jameson 1", d, 1, source)


def williams_shunn_jameson_2():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 6)]]}
    return T2Scheme("Williams-Shunn-Jameson 2", d, 2, source)


def williams_shunn_jameson_3():
    return _read(this_dir / "williams_shunn_jameson_3.json", source)


def williams_shunn_jameson_4():
    return _read(this_dir / "williams_shunn_jameson_4.json", source)


def williams_shunn_jameson_5():
    return _read(this_dir / "williams_shunn_jameson_5.json", source)


def williams_shunn_jameson_6():
    return _read(this_dir / "williams_shunn_jameson_6.json", source)


def williams_shunn_jameson_7():
    return _read(this_dir / "williams_shunn_jameson_7.json", source)


def williams_shunn_jameson_8():
    return _read(this_dir / "williams_shunn_jameson_8.json", source)


register(
    [
        williams_shunn_jameson_1,
        williams_shunn_jameson_2,
        williams_shunn_jameson_3,
        williams_shunn_jameson_4,
        williams_shunn_jameson_5,
        williams_shunn_jameson_6,
        williams_shunn_jameson_7,
        williams_shunn_jameson_8,
    ]
)



from ._cools_haegemans import cools_haegemans_1, cools_haegemans_2
from ._dobrodeev_1970 import dobrodeev_1970
from ._dobrodeev_1978 import dobrodeev_1978
from ._ewing import ewing
from ._hammer_stroud import hammer_stroud_1n, hammer_stroud_2n
from ._helpers import CnScheme, ncube_points, transform
from ._mcnamee_stenger import (
    mcnamee_stenger_3,
    mcnamee_stenger_5,
    mcnamee_stenger_7a,
    mcnamee_stenger_7b,
    mcnamee_stenger_9a,
    mcnamee_stenger_9b,
)
from ._mustard_lyness_blatt import mustard_lyness_blatt
from ._phillips import phillips
from ._stroud import (
    stroud_cn_1_1,
    stroud_cn_1_2,
    stroud_cn_2_1,
    stroud_cn_2_2,
    stroud_cn_3_1,
    stroud_cn_3_2,
    stroud_cn_3_3,
    stroud_cn_3_4,
    stroud_cn_3_5,
    stroud_cn_3_6,
    stroud_cn_5_2,
    stroud_cn_5_3,
    stroud_cn_5_4,
    stroud_cn_5_5,
    stroud_cn_5_6,
    stroud_cn_5_7,
    stroud_cn_5_8,
    stroud_cn_5_9,
    stroud_cn_7_1,
)
from ._stroud_1957 import stroud_1957_2, stroud_1957_3
from ._stroud_1966 import stroud_1966_a, stroud_1966_b, stroud_1966_c, stroud_1966_d
from ._stroud_1968 import stroud_1968
from ._thacher import thacher
from ._tyler import tyler

__all__ = [
    "cools_haegemans_1",
    "cools_haegemans_2",
    "dobrodeev_1970",
    "dobrodeev_1978",
    "ewing",
    "hammer_stroud_1n",
    "hammer_stroud_2n",
    "mcnamee_stenger_3",
    "mcnamee_stenger_5",
    "mcnamee_stenger_7a",
    "mcnamee_stenger_7b",
    "mcnamee_stenger_9a",
    "mcnamee_stenger_9b",
    "mustard_lyness_blatt",
    "phillips",
    "stroud_1957_2",
    "stroud_1957_3",
    "stroud_1966_a",
    "stroud_1966_b",
    "stroud_1966_c",
    "stroud_1966_d",
    "stroud_1968",
    "stroud_cn_1_1",
    "stroud_cn_1_2",
    "stroud_cn_2_1",
    "stroud_cn_2_2",
    "stroud_cn_3_1",
    "stroud_cn_3_2",
    "stroud_cn_3_3",
    "stroud_cn_3_4",
    "stroud_cn_3_5",
    "stroud_cn_3_6",
    "stroud_cn_5_2",
    "stroud_cn_5_3",
    "stroud_cn_5_4",
    "stroud_cn_5_5",
    "stroud_cn_5_6",
    "stroud_cn_5_7",
    "stroud_cn_5_8",
    "stroud_cn_5_9",
    "stroud_cn_7_1",
    "thacher",
    "tyler",
    #
    "ncube_points",
    "transform",
    "CnScheme",
]



import numpy as np

from ..helpers import techreport
from ._helpers import C2Scheme, register

source = techreport(
    authors=["Shayne Waldron"],
    title="Symmetries of linear functionals",
    institution="University of Wisconsin-Madison, Center for Mathematical Sciences",
    note="CMS Technical Summary Report",
    month="oct",
    year="1994",
    url="https://www.semanticscholar.org/paper/Symmetries-of-Linear-Functionals-Waldron/086b084bac739c4616a06cc503dfcbda9f3cc7e8",
)


def waldron(r=1, alpha=0):
    assert r ** 2 >= 1 / 3

    R = r / np.sqrt(3 * r ** 2 - 1)

    beta = alpha + np.pi / 2

    sin_alpha = np.sin(alpha)
    cos_alpha = np.cos(alpha)
    sin_beta = np.sin(beta)
    cos_beta = np.cos(beta)

    d = {
        "plain": [
            [1 / 6 / r ** 2, 1 / 6 / r ** 2, 1 / 6 / R ** 2, 1 / 6 / R ** 2],
            [+r * cos_alpha, -r * cos_alpha, +R * cos_beta, -R * cos_beta],
            [+r * sin_alpha, -r * sin_alpha, +R * sin_beta, -R * sin_beta],
        ]
    }
    return C2Scheme("Waldron", d, 3, source)


register([waldron])



from . import enr, enr2, nball, ncube, nsimplex, nsphere
from .__about__ import __version__

__all__ = [
    "__version__",
    "enr",
    "enr2",
    "nball",
    "ncube",
    "nsimplex",
    "nsphere",
]



import numpy as np
import orthopy

from ..tools import scheme_from_rc
from ._helpers import E1r2Scheme


def gauss_hermite(n, mode="numpy"):
    """
    Gauss-Hermite quadrature for integrals of the form

        int_{-inf}^{+inf} exp(-x^2) f(x) dx.
    """
    if mode == "numpy":
        points, weights = np.polynomial.hermite.hermgauss(n)
    else:
        rc = orthopy.e1r2.RecurrenceCoefficients("physicists", "monic", symbolic=True)
        _, alpha, beta = np.array([rc[k] for k in range(n)]).T
        points, weights = scheme_from_rc(alpha, beta, rc.int_1, mode=mode)
    return E1r2Scheme(f"Gauss-Hermite ({n})", weights, points, 2 * n - 1)



import numpy as np
import sympy

from ..helpers import article
from ._helpers import S2Scheme, register

_source = article(
    authors=["I.P. Mysovskikh"],
    title="On the construction of cubature formulas for the simplest regions",
    journal="Z. Vychisl. Mat. i. Mat. Fiz.",
    number="4",
    pages="3-14",
    year="1964",
)

frac = sympy.Rational
pi = sympy.pi
sqrt = np.vectorize(sympy.sqrt)
cos = np.vectorize(sympy.cos)
sin = np.vectorize(sympy.sin)
pm_ = np.array([+1, -1])


def mysovskih_1(alpha=0):
    b = sqrt(frac(alpha + 4, alpha + 6))

    B0 = frac(4, (alpha + 4) ** 2)
    B1 = frac((alpha + 2) * (alpha + 6), 5 * (alpha + 4) ** 2)

    d = {
        "zero2": [[B0]],
        "d5.0": [[B1], [b]],
    }
    return S2Scheme("Mysovskih 1", d, 4, _source)


def mysovskih_2():
    sqrt10 = sqrt(10)
    sqrt601 = sqrt(601)

    B1, B3 = (857 * sqrt601 + pm_ * 12707) / 20736 / sqrt601
    B2 = frac(125, 3456)
    B4, B5 = (340 + pm_ * 25 * sqrt10) / 10368

    r1, r3 = sqrt((31 - pm_ * sqrt601) / 60)
    r2 = sqrt(frac(3, 5))
    r4, r5 = sqrt((10 - pm_ * sqrt10) / 20)

    c4, s5 = sqrt((10 - pm_ * sqrt10) / 60)

    d = {"d4_a0": [[B1, B2, B3], [r1, r2, r3]], "d4_ab": [[B4, B5], [r4, r5], [c4, s5]]}
    return S2Scheme("Mysovskih 2", d, 11, _source)


def mysovskih_3():
    sqrt21 = sqrt(21)
    sqrt1401 = sqrt(1401)

    # ERR Stroud lists 5096 instead of 4998 here
    A1, A2 = (4998 + pm_ * 343 * sqrt21) / 253125

    # ERR Stroud is missing the +- here
    B1, B3 = (1055603 * sqrt1401 + pm_ * 26076047) / 43200000 / sqrt1401

    B2 = frac(16807, 800000)

    rho1, rho2 = sqrt((21 - pm_ * sqrt21) / 28)
    sigma1, sigma3 = sqrt((69 - pm_ * sqrt1401) / 112)
    sigma2 = sqrt(frac(5, 7))

    tau1 = 0.252863797091230
    tau2 = 0.577728928444823
    tau3 = 0.873836956644882
    tau4 = 0.989746802511491

    C1 = 0.398811120280412e-1
    C2 = 0.348550570365141e-1
    C3 = 0.210840370156484e-1
    C4 = 0.531979391979623e-2

    d = {
        "d8.1": [[A1, A2], [rho1, rho2]],
        "d4.1": [[B1, B2, B3], [sigma1, sigma2, sigma3]],
        "d4.0": [[C1, C2, C3, C4], [tau1, tau2, tau3, tau4]],
    }
    return S2Scheme("Mysovskih 3", d, 15, _source)


register([mysovskih_1, mysovskih_2, mysovskih_3])



import math

import numpy as np
import sympy

from ..helpers import book, fsd, pm, untangle
from ._helpers import Enr2Scheme
from ._stenger import stenger_7a as stroud_enr2_7_3a
from ._stenger import stenger_7b as stroud_enr2_7_3b
from ._stenger import stenger_9a as stroud_enr2_9_1a
from ._stenger import stenger_9b as stroud_enr2_9_1b
from ._stenger import stenger_11a as stroud_enr2_11_1a
from ._stenger import stenger_11b as stroud_enr2_11_1b
from ._stroud_1967_5 import stroud_1967_5_a as stroud_enr2_5_1a
from ._stroud_1967_5 import stroud_1967_5_b as stroud_enr2_5_1b
from ._stroud_1967_7 import stroud_1967_7_2a as stroud_enr2_7_1a
from ._stroud_1967_7 import stroud_1967_7_2b as stroud_enr2_7_1b
from ._stroud_1967_7 import stroud_1967_7_4 as stroud_enr2_7_2
from ._stroud_secrest import stroud_secrest_1 as stroud_enr2_3_1
from ._stroud_secrest import stroud_secrest_2 as stroud_enr2_3_2
from ._stroud_secrest import stroud_secrest_4 as stroud_enr2_5_2

source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_enr2_5_3(n, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    assert n > 2

    r = sqrt(frac(n + 2, 4))
    s = sqrt(frac(n + 2, 2 * (n - 2)))
    A = frac(4, (n + 2) ** 2)
    B = frac((n - 2) ** 2, 2 ** n * (n + 2) ** 2)

    data = [(A, fsd(n, (r, 1))), (B, pm(n * [s]))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Stroud Enr2 5-3", n, weights, points, 5, source)


def stroud_enr2_5_4(n, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    # Spherical product Lobatto
    B0 = frac(2, (n + 2))
    data = [(B0, [n * [0]])]
    for k in range(1, n + 1):
        rk = sqrt(frac(k + 2, 2))
        s = sqrt(frac(1, 2))
        arr = (k - 1) * [0] + [rk] + (n - k) * [s]
        alpha = frac(2 ** (k - n), (k + 1) * (k + 2))
        data += [(alpha, pm(arr))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Stroud Enr2 5-4", n, weights, points, 5, source)


def _stroud_5_5(n, variant_a, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    p_m = +1 if variant_a else -1

    # r is complex-valued for n >= 3
    r = sqrt((n + 2 + p_m * (n - 1) * sqrt(2 * (n + 2))) / (2 * n))
    s = sqrt((n + 2 - p_m * sqrt(2 * (n + 2))) / (2 * n))
    A = frac(2, n + 2)
    B = frac(1, 2 ** n * (n + 2))

    data = [(A, [n * [0]]), (B, fsd(n, (r, 1), (s, n - 1)))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    variant = "a" if variant_a else "b"
    return Enr2Scheme(f"Stroud Enr2 5-5{variant}", n, weights, points, 5, source)


def stroud_enr2_5_5a(n):
    return _stroud_5_5(n, True)


def stroud_enr2_5_5b(n):
    return _stroud_5_5(n, False)


def stroud_enr2_5_6(n, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    assert n >= 5

    sqrt2 = sqrt(2)
    sqrt2n1 = sqrt(2 * (n + 1))
    r = sqrt((n - sqrt2 + (n - 1) * sqrt2n1) / (2 * n))
    s = sqrt((n - sqrt2 - sqrt2n1) / (2 * n))
    t = sqrt((1 + sqrt2) / 2)
    A = frac(1, 2 ** n * (n + 1))

    data = [(A, fsd(n, (r, 1), (s, n - 1))), (A, pm(n * [t]))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Stroud Enr2 5-6", n, weights, points, 5, source)


__all__ = [
    "stroud_enr2_3_1",
    "stroud_enr2_3_2",
    "stroud_enr2_5_1a",
    "stroud_enr2_5_1b",
    "stroud_enr2_5_2",
    "stroud_enr2_5_3",
    "stroud_enr2_5_4",
    "stroud_enr2_5_5a",
    "stroud_enr2_5_5b",
    "stroud_enr2_5_6",
    "stroud_enr2_7_1a",
    "stroud_enr2_7_1b",
    "stroud_enr2_7_2",
    "stroud_enr2_7_3a",
    "stroud_enr2_7_3b",
    "stroud_enr2_9_1a",
    "stroud_enr2_9_1b",
    "stroud_enr2_11_1a",
    "stroud_enr2_11_1b",
]



from __future__ import annotations

import numpy as np
import orthopy

from ..tools import scheme_from_rc
from ._helpers import E1rScheme


def gauss_laguerre(n: int, alpha: int | float = 0, mode: str = "numpy"):
    """
    Gauss-Laguerre quadrature for integrals of the form

        int_0^{+inf} exp(-alpha * x) f(x) dx.
    """
    symbolic = mode != "numpy"
    rc = orthopy.e1r.RecurrenceCoefficients("monic", alpha, symbolic)

    _, a, b = np.array([rc[k] for k in range(n)]).T
    points, weights = scheme_from_rc(a, b, rc.int_1, mode=mode)

    if alpha == 0:
        name = f"Gauss-Laguerre ({n})"
    else:
        name = f"Generalized Gauss-Laguerre (n={n}, alpha={alpha})"

    return E1rScheme(name, weights, points, 2 * n - 1)



"""
Arthur Stroud,
Approximate Calculation of Multiple Integrals,
Prentice Hall, 1971.
"""
from ._stroud_secrest import stroud_secrest_07 as stroud_e3r_5_1
from ._stroud_secrest import stroud_secrest_08 as stroud_e3r_5_2
from ._stroud_secrest import stroud_secrest_09 as stroud_e3r_5_3
from ._stroud_secrest import stroud_secrest_10 as stroud_e3r_7_1
from ._stroud_secrest import stroud_secrest_11 as stroud_e3r_7_2

__all__ = [
    "stroud_e3r_5_1",
    "stroud_e3r_5_2",
    "stroud_e3r_5_3",
    "stroud_e3r_7_1",
    "stroud_e3r_7_2",
]



from sympy import Rational as frac

from ..helpers import article
from ._helpers import C3Scheme, register

_source = article(
    authors=["J. Albrecht", "L. Collatz"],
    title="Zur numerischen Auswertung mehrdimensionaler Integrale",
    journal="ZAMM",
    volume="38",
    number="1-2",
    year="1958",
    pages="1–15",
    url="https://doi.org/10.1002/zamm.19580380102",
)


def albrecht_collatz():
    d = {
        "zero3": [[frac(1, 4)]],
        "symm_r00": [[frac(1, 12)], [1]],
        "symm_rr0": [[frac(1, 48)], [1]],
    }
    return C3Scheme("Albrecht-Collatz", d, 3, _source)


register([albrecht_collatz])



import json

import numpy as np
import orthopy
import sympy

from .._exception import QuadpyError
from ..helpers import QuadratureScheme

schemes = {}


def register(in_schemes):
    for scheme in in_schemes:
        schemes[scheme.__name__] = scheme


class U3Scheme(QuadratureScheme):
    def __init__(
        self,
        name,
        symmetry_data,
        degree,
        source,
        tol=1.0e-14,
        comments=None,
    ):
        self.symmetry_data = symmetry_data
        points, weights = expand_symmetries(symmetry_data)

        if np.asarray(points).dtype == sympy.Basic:
            theta_phi = cartesian_to_spherical_sympy(points)
        else:
            theta_phi = cartesian_to_spherical(points)

        self.domain = "U3"
        super().__init__(name, weights, points, degree, source, tol, comments)

        theta_phi = np.asarray(theta_phi)
        if theta_phi.dtype == np.float64:
            self.theta_phi = theta_phi
        else:
            assert theta_phi.dtype in [np.dtype("O"), np.int_]
            self.theta_phi = theta_phi.astype(np.float64)
            self.theta_phi_symbolic = theta_phi

    def plot(self):
        from matplotlib import pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        ax = plt.axes(projection=Axes3D.name)
        # ax.set_aspect("equal")

        for p, w in zip(self.points.T, self.weights):
            # <https://en.wikipedia.org/wiki/Spherical_cap>
            w *= 4 * np.pi
            theta = np.arccos(1.0 - abs(w) / (2 * np.pi))
            color = "tab:blue" if w >= 0 else "tab:red"
            _plot_spherical_cap_mpl(ax, p, theta, color)

        ax.set_axis_off()
        return plt

    def integrate(self, f, center, radius, dot=np.dot):
        """Quadrature where `f` is defined in Cartesian coordinates."""
        center = np.asarray(center)
        rr = (self.points.T * radius + center).T
        frr = f(rr)
        if frr.shape[-len(rr.shape[1:]) :] != rr.shape[1:]:
            string = ", ".join(str(val) for val in rr.shape[1:])
            raise QuadpyError(
                f"Wrong return value shape {frr.shape}. Expected (..., {string})."
            )
        return area(radius) * dot(frr, self.weights)

    def integrate_spherical(self, f, dot=np.dot):
        """Quadrature where `f` is a function of the spherical coordinates theta_phi
        (polar, azimuthal, in this order).
        """
        ff = f(self.theta_phi)
        return area(1.0) * dot(ff, self.weights)

    def compute_residuals(self, level):
        evaluator = orthopy.u3.EvalCartesian(self.points, "quantum mechanic")

        max_res = []
        for k in range(level + 1):
            approximate = self.integrate(
                lambda x: next(evaluator), [0.0, 0.0, 0.0], 1.0
            )
            exact = evaluator.int_p0 if k == 0 else 0.0
            res = np.abs(approximate - exact)
            max_res += [np.max(res)]

        return np.array(max_res)


def area(radius):
    return 4 * np.pi * np.array(radius) ** 2


def _plot_spherical_cap_mpl(ax, b, opening_angle, color, elevation=1.01):
    r = elevation
    azimuthal = np.linspace(0, 2 * np.pi, 30)
    polar = np.linspace(0, opening_angle, 20)
    X = r * np.stack(
        [
            np.outer(np.cos(azimuthal), np.sin(polar)),
            np.outer(np.sin(azimuthal), np.sin(polar)),
            np.outer(np.ones(np.size(azimuthal)), np.cos(polar)),
        ],
        axis=-1,
    )

    # rotate X such that [0, 0, 1] gets rotated to `c`;
    # <https://math.stackexchange.com/a/476311/36678>.
    a = np.array([0.0, 0.0, 1.0])
    a_x_b = np.cross(a, b)
    a_dot_b = np.dot(a, b)
    if a_dot_b == -1.0:
        X_rot = -X
    else:
        X_rot = (
            X
            + np.cross(a_x_b, X)
            + np.cross(a_x_b, np.cross(a_x_b, X)) / (1.0 + a_dot_b)
        )

    ax.plot_surface(
        X_rot[..., 0],
        X_rot[..., 1],
        X_rot[..., 2],
        rstride=3,
        cstride=3,
        color=color,
        alpha=0.5,
        linewidth=0,
    )


def cartesian_to_spherical(X):
    return np.array([np.arccos(X[2]), np.arctan2(X[1], X[0])])


def _atan2_0(X):
    """Like sympy.atan2, but return 0 for x=y=0. Mathematically, the value is
    undefined, so sympy returns NaN, but for the sake of the coordinate
    conversion, its value doesn't matter. NaNs, however, produce NaNs down the
    line.
    """
    out = np.array([sympy.atan2(X[1, k], X[0, k]) for k in range(X.shape[1])])
    out[out == sympy.nan] = 0
    return out


def cartesian_to_spherical_sympy(X):
    arccos = np.vectorize(sympy.acos)
    return np.array([arccos(X[2]), _atan2_0(X)])


def _a1(vals):
    symbolic = np.asarray(vals).dtype == sympy.Basic
    a = 1 if symbolic else 1.0
    points = np.array(
        [[+a, 0, 0], [-a, 0, 0], [0, +a, 0], [0, -a, 0], [0, 0, +a], [0, 0, -a]]
    ).T
    return points


def _a2(vals):
    symbolic = np.asarray(vals).dtype == sympy.Basic
    a = 1 / sympy.sqrt(2) if symbolic else 1 / np.sqrt(2)
    points = np.array(
        [
            [+a, +a, 0],
            [+a, -a, 0],
            [-a, +a, 0],
            [-a, -a, 0],
            #
            [+a, 0, +a],
            [+a, 0, -a],
            [-a, 0, +a],
            [-a, 0, -a],
            #
            [0, +a, +a],
            [0, +a, -a],
            [0, -a, +a],
            [0, -a, -a],
        ]
    ).T
    return points


def _a3(vals):
    symbolic = np.asarray(vals).dtype == sympy.Basic
    a = 1 / sympy.sqrt(3) if symbolic else 1 / np.sqrt(3)
    points = np.array(
        [
            [+a, +a, +a],
            [+a, +a, -a],
            [+a, -a, +a],
            [+a, -a, -a],
            [-a, +a, +a],
            [-a, +a, -a],
            [-a, -a, +a],
            [-a, -a, -a],
        ]
    ).T
    return points


def _pq0(vals):
    return _pq02([np.sin(vals[0] * np.pi), np.cos(vals[0] * np.pi)])


def _pq02(vals):
    if len(vals) == 1:
        a = vals[0]
        b = np.sqrt(1 - a ** 2)
    else:
        assert len(vals) == 2
        a, b = vals

    if isinstance(a, sympy.Basic):
        zero = 0
    else:
        zero = np.zeros_like(a)

    points = np.array(
        [
            [+a, +b, zero],
            [-a, +b, zero],
            [-a, -b, zero],
            [+a, -b, zero],
            #
            [+b, +a, zero],
            [-b, +a, zero],
            [-b, -a, zero],
            [+b, -a, zero],
            #
            [+a, zero, +b],
            [-a, zero, +b],
            [-a, zero, -b],
            [+a, zero, -b],
            #
            [+b, zero, +a],
            [-b, zero, +a],
            [-b, zero, -a],
            [+b, zero, -a],
            #
            [zero, +a, +b],
            [zero, -a, +b],
            [zero, -a, -b],
            [zero, +a, -b],
            #
            [zero, +b, +a],
            [zero, -b, +a],
            [zero, -b, -a],
            [zero, +b, -a],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _rs0(vals):
    if len(vals) == 1:
        a = vals[0]
        b = np.sqrt(1 - a ** 2)
    else:
        assert len(vals) == 2
        a, b = vals

    if isinstance(a, sympy.Basic):
        zero = 0
    else:
        zero = np.zeros_like(a)

    points = np.array(
        [
            [+a, +b, zero],
            [-a, +b, zero],
            [-a, -b, zero],
            [+a, -b, zero],
            #
            [+b, zero, +a],
            [-b, zero, +a],
            [-b, zero, -a],
            [+b, zero, -a],
            #
            [zero, +a, +b],
            [zero, -a, +b],
            [zero, -a, -b],
            [zero, +a, -b],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _llm(vals):
    # translate the point into cartesian coords; note that phi=pi/4.
    beta = vals[0] * np.pi
    L = np.sin(beta) / np.sqrt(2)
    m = np.cos(beta)
    return _llm2([L, m])


def _llm2(vals):
    if len(vals) == 1:
        L = vals[0]
        m = np.sqrt(1 - 2 * L ** 2)
    else:
        assert len(vals) == 2
        L, m = vals

    points = np.array(
        [
            [+L, +L, +m],
            [-L, +L, +m],
            [+L, -L, +m],
            [-L, -L, +m],
            [+L, +L, -m],
            [-L, +L, -m],
            [+L, -L, -m],
            [-L, -L, -m],
            #
            [+L, +m, +L],
            [-L, +m, +L],
            [+L, +m, -L],
            [-L, +m, -L],
            [+L, -m, +L],
            [-L, -m, +L],
            [+L, -m, -L],
            [-L, -m, -L],
            #
            [+m, +L, +L],
            [+m, -L, +L],
            [+m, +L, -L],
            [+m, -L, -L],
            [-m, +L, +L],
            [-m, -L, +L],
            [-m, +L, -L],
            [-m, -L, -L],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _rsw(vals):
    # translate the point into cartesian coords; note that phi=pi/4.
    phi_theta = vals * np.pi

    sin_phi, sin_theta = np.sin(phi_theta)
    cos_phi, cos_theta = np.cos(phi_theta)

    return _rsw2([sin_theta * cos_phi, sin_theta * sin_phi, cos_theta])


def _rsw2(vals):
    if len(vals) == 2:
        r, s = vals
        w = np.sqrt(1 - r ** 2 - s ** 2)
    else:
        assert len(vals) == 3
        r, s, w = vals

    points = np.array(
        [
            [+r, +s, +w],
            [+w, +r, +s],
            [+s, +w, +r],
            [+s, +r, +w],
            [+w, +s, +r],
            [+r, +w, +s],
            #
            [-r, +s, +w],
            [+w, -r, +s],
            [+s, +w, -r],
            [+s, -r, +w],
            [+w, +s, -r],
            [-r, +w, +s],
            #
            [+r, -s, +w],
            [+w, +r, -s],
            [-s, +w, +r],
            [-s, +r, +w],
            [+w, -s, +r],
            [+r, +w, -s],
            #
            [+r, +s, -w],
            [-w, +r, +s],
            [+s, -w, +r],
            [+s, +r, -w],
            [-w, +s, +r],
            [+r, -w, +s],
            #
            [-r, -s, +w],
            [+w, -r, -s],
            [-s, +w, -r],
            [-s, -r, +w],
            [+w, -s, -r],
            [-r, +w, -s],
            #
            [-r, +s, -w],
            [-w, -r, +s],
            [+s, -w, -r],
            [+s, -r, -w],
            [-w, +s, -r],
            [-r, -w, +s],
            #
            [+r, -s, -w],
            [-w, +r, -s],
            [-s, -w, +r],
            [-s, +r, -w],
            [-w, -s, +r],
            [+r, -w, -s],
            #
            [-r, -s, -w],
            [-w, -r, -s],
            [-s, -w, -r],
            [-s, -r, -w],
            [-w, -s, -r],
            [-r, -w, -s],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _rst(vals):
    if len(vals) == 2:
        r, s = vals
        w = np.sqrt(1 - r ** 2 - s ** 2)
    else:
        assert len(vals) == 3
        r, s, w = vals

    points = np.array(
        [
            [+r, +s, +w],
            [+w, +r, +s],
            [+s, +w, +r],
            #
            [-r, +s, +w],
            [+w, -r, +s],
            [+s, +w, -r],
            #
            [+r, -s, +w],
            [+w, +r, -s],
            [-s, +w, +r],
            #
            [+r, +s, -w],
            [-w, +r, +s],
            [+s, -w, +r],
            #
            [-r, -s, +w],
            [+w, -r, -s],
            [-s, +w, -r],
            #
            [-r, +s, -w],
            [-w, -r, +s],
            [+s, -w, -r],
            #
            [+r, -s, -w],
            [-w, +r, -s],
            [-s, -w, +r],
            #
            [-r, -s, -w],
            [-w, -r, -s],
            [-s, -w, -r],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _rst_weird(vals):
    if len(vals) == 2:
        r, s = vals
        t = np.sqrt(1 - r ** 2 - s ** 2)
    else:
        assert len(vals) == 3
        r, s, t = vals

    points = np.array(
        [
            [+r, +s, +t],
            [-r, +t, +s],
            [+s, +t, +r],
            [-s, +r, +t],
            [+t, +r, +s],
            [-t, +s, +r],
            #
            [+r, -s, -t],
            [-r, -t, -s],
            [+s, -t, -r],
            [-s, -r, -t],
            [+t, -r, -s],
            [-t, -s, -r],
            #
            [+r, +t, -s],
            [-r, +s, -t],
            [+s, +r, -t],
            [-s, +t, -r],
            [+t, +s, -r],
            [-t, +r, -s],
            #
            [+r, -t, +s],
            [-r, -s, +t],
            [+s, -r, +t],
            [-s, -t, +r],
            [+t, -s, +r],
            [-t, -r, +s],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def expand_symmetries_points_only(data):
    points = []
    counts = []

    for key, points_raw in data.items():
        fun = {
            "a1": _a1,
            "a2": _a2,
            "a3": _a3,
            "llm": _llm,
            "llm2": _llm2,
            "pq0": _pq0,
            "pq02": _pq02,
            "rs0": _rs0,
            "rsw": _rsw,
            "rsw2": _rsw2,
            "rst": _rst,
            "rst_weird": _rst_weird,
            "plain": lambda vals: vals.reshape(3, 1, -1),
        }[key]
        pts = fun(np.asarray(points_raw))

        counts.append(pts.shape[1])
        pts = pts.reshape(pts.shape[0], -1)
        points.append(pts)

    points = np.ascontiguousarray(np.concatenate(points, axis=1))
    return points, counts


def expand_symmetries(data):
    # separate points and weights
    points_raw = {}
    weights_raw = []
    for key, values in data.items():
        weights_raw.append(values[0])
        points_raw[key] = values[1:]

    points, counts = expand_symmetries_points_only(points_raw)
    weights = np.concatenate(
        [np.tile(values, count) for count, values in zip(counts, weights_raw)]
    )
    return points, weights


def _scheme_from_dict(content, source=None):
    data = content["data"]
    if "weight factor" in content:
        w = content["weight factor"]
        for val in data.values():
            val[0] = [v * w for v in val[0]]

    return U3Scheme(
        content["name"],
        data,
        degree=content["degree"],
        source=source,
        tol=content["test_tolerance"],
        comments=content["comments"] if "comments" in content else None,
    )


def _read(filepath, source):
    with open(filepath) as f:
        content = json.load(f)
    return _scheme_from_dict(content, source)


def get_good_scheme(degree):
    if degree <= 47:
        return {
            0: schemes["lebedev_003a"],
            1: schemes["lebedev_003a"],
            2: schemes["lebedev_003a"],
            3: schemes["lebedev_003a"],
            4: schemes["albrecht_collatz_1"],
            5: schemes["albrecht_collatz_1"],
            6: schemes["albrecht_collatz_5"],
            7: schemes["albrecht_collatz_5"],
            8: schemes["mclaren_05"],
            9: schemes["mclaren_05"],
            10: schemes["mclaren_08"],
            11: schemes["mclaren_08"],
            12: schemes["lebedev_015"],
            13: schemes["lebedev_015"],
            14: schemes["lebedev_015"],
            15: schemes["lebedev_015"],
            16: schemes["lebedev_017"],
            17: schemes["lebedev_017"],
            18: schemes["lebedev_019"],
            19: schemes["lebedev_019"],
            20: schemes["lebedev_021"],
            21: schemes["lebedev_021"],
            22: schemes["lebedev_023"],
            23: schemes["lebedev_023"],
            24: schemes["heo_xu_25b"],
            25: schemes["heo_xu_25b"],
            26: schemes["lebedev_029"],
            27: schemes["lebedev_029"],
            28: schemes["lebedev_029"],
            29: schemes["lebedev_029"],
            30: schemes["lebedev_031"],
            31: schemes["lebedev_031"],
            32: schemes["lebedev_035"],
            33: schemes["lebedev_035"],
            34: schemes["lebedev_035"],
            35: schemes["lebedev_035"],
            36: schemes["lebedev_041"],
            37: schemes["lebedev_041"],
            38: schemes["lebedev_041"],
            39: schemes["lebedev_041"],
            40: schemes["lebedev_041"],
            41: schemes["lebedev_041"],
            42: schemes["lebedev_047"],
            43: schemes["lebedev_047"],
            44: schemes["lebedev_047"],
            45: schemes["lebedev_047"],
            46: schemes["lebedev_047"],
            47: schemes["lebedev_047"],
        }[degree]()

    return None



"""
Two of the schemes also appear in

P.C. Hammer, Arthur H. Stroud,
Numerical Evaluation of Multiple Integrals II,
Mathematical Tables and Other Aids to Computation.
Vol. 12, No. 64 (Oct., 1958), pp. 272-280,
<https://www.jstor.org/stable/2002370>
"""
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import T3Scheme, register

source = article(
    authors=["P.C. Hammer", "O.J. Marlowe", "A.H. Stroud"],
    title="Numerical Integration Over Simplexes and Cones",
    journal="Mathematical Tables and Other Aids to Computation",
    volume="10",
    number="55",
    month="jul",
    year="1956",
    pages="130-137",
    url="https://doi.org/10.1090/S0025-5718-1956-0086389-6",
)
# source = book(
#     authors=["Olgierd Zienkiewicz"],
#     title="The Finite Element Method, Sixth Edition",
#     publisher="Butterworth-Heinemann",
#     year="2005",
#     isbn="0750663200",
#     url="https://www.sciencedirect.com/book/9780750664318/the-finite-element-method-set",
# )


# Used in Zienkiewicz 4
def hammer_marlowe_stroud_1():
    degree = 2
    r = 1 / sqrt(5)
    a = (1 - r) / 4
    d = {"s31": [[frac(1, 4)], [a]]}
    return T3Scheme("Hammer-Marlowe-Stroud 1", d, degree, source)


def hammer_marlowe_stroud_2():
    degree = 2
    r = -1 / sqrt(5)
    a = (1 - r) / 4
    d = {"s31": [[frac(1, 4)], [a]]}
    return T3Scheme("Hammer-Marlowe-Stroud 2", d, degree, source)


# Used in Zienkiewicz 5
def hammer_marlowe_stroud_3():
    degree = 3
    r = frac(1, 3)
    a = (1 - r) / 4
    d = {"s4": [[-frac(4, 5)]], "s31": [[frac(9, 20)], [a]]}
    return T3Scheme("Hammer-Marlowe-Stroud 3", d, degree, source)


register([hammer_marlowe_stroud_1, hammer_marlowe_stroud_2, hammer_marlowe_stroud_3])



import numpy as np
import orthopy

from ..tools import scheme_from_rc
from ._helpers import C1Scheme


def gauss_lobatto(n, a=0.0, b=0.0):
    assert n >= 2
    degree = 2 * n - 3
    rc = orthopy.c1.jacobi.RecurrenceCoefficients("monic", a, b, symbolic=False)
    _, alpha, beta = np.array([rc[k] for k in range(n)]).T
    beta[0] = rc.int_1
    points, weights = _lobatto(alpha, beta, rc.int_1, -1.0, 1.0)
    return C1Scheme("Gauss-Lobatto", degree, weights, points)


def _lobatto(alpha, beta, int_1, xl1, xl2):
    """Compute the Lobatto nodes and weights with the preassigned node xl1, xl2.
    Based on the section 7 of the paper

        Some modified matrix eigenvalue problems,
        Gene Golub,
        SIAM Review Vol 15, No. 2, April 1973, pp.318--334,

    and

        http://www.scientificpython.net/pyblog/radau-quadrature
    """
    from scipy.linalg import solve, solve_banded

    n = len(alpha) - 1
    en = np.zeros(n)
    en[-1] = 1
    A1 = np.vstack((np.sqrt(beta), alpha - xl1))
    J1 = np.vstack((A1[:, 0:-1], A1[0, 1:]))
    A2 = np.vstack((np.sqrt(beta), alpha - xl2))
    J2 = np.vstack((A2[:, 0:-1], A2[0, 1:]))
    g1 = solve_banded((1, 1), J1, en)
    g2 = solve_banded((1, 1), J2, en)
    C = np.array(((1, -g1[-1]), (1, -g2[-1])))
    xl = np.array((xl1, xl2))
    ab = solve(C, xl)

    alphal = alpha.copy()
    alphal[-1] = ab[0]
    betal = beta.copy()
    betal[-1] = ab[1]
    x, w = scheme_from_rc(alphal, betal, int_1, mode="numpy")
    return x, w



import pathlib

from ...helpers import article, techreport
from .._helpers import _read, register

source = techreport(
    authors=["J. Berntsen", "T.O. Espelid"],
    title="Degree 13 symmetric quadrature rules for the triangle",
    # Reports in Informatics,
    institution="Dept. of Informatics, University of Bergen",
    year="1990",
)

# This first scheme was published separately as
c2 = article(
    authors=["J. Berntsen", "T.O. Espelid"],
    title="Algorithm 706: DCUTRI: An Algorithm for Adaptive Cubature over a Collection of Triangles",
    journal="ACM Trans. Math. Softw.",
    month="sep",
    year="1992",
    url="https://doi.org/10.1145/131766.131772",
)

this_dir = pathlib.Path(__file__).resolve().parent


def dcutri():
    out = berntsen_espelid_1()
    out.source = c2
    return out


def berntsen_espelid_1():
    return _read(this_dir / "berntsen_espelid_1.json", source)


def berntsen_espelid_2():
    return _read(this_dir / "berntsen_espelid_2.json", source)


def berntsen_espelid_3():
    return _read(this_dir / "berntsen_espelid_3.json", source)


def berntsen_espelid_4():
    return _read(this_dir / "berntsen_espelid_4.json", source)


register(
    [
        dcutri,
        berntsen_espelid_1,
        berntsen_espelid_2,
        berntsen_espelid_3,
        berntsen_espelid_4,
    ]
)



from . import (
    _albrecht_collatz,
    _bazant_oh,
    _fliege_maier,
    _heo_xu,
    _lebedev,
    _mclaren,
    _stroud,
)
from ._helpers import area, get_good_scheme, schemes

__all__ = [
    "_albrecht_collatz",
    "_bazant_oh",
    "_fliege_maier",
    "_heo_xu",
    "_lebedev",
    "_mclaren",
    "_stroud",
    #
    "area",
    "schemes",
    "get_good_scheme",
]



from sympy import Rational as frac

from ..helpers import article, expand_symmetries
from ._helpers import CnScheme

_source = article(
    authors=["G.M. Ewing"],
    title="On Approximate Cubature",
    journal="The American Mathematical Monthly",
    volume="48",
    number="2",
    month="feb",
    year="1941",
    pages="134-136",
    url="https://doi.org/10.2307/2303604",
)


def ewing(n):
    d = {"0": [[frac(2, 3)]], "a": [[frac(1, 3 * 2 ** n)], [1]]}
    points, weights = expand_symmetries(d, n)
    return CnScheme("Ewing", n, weights, points, 3, _source)



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import C3Scheme, register

source = article(
    authors=["Michael Sadowsky"],
    title="A Formula for Approximate Computation of a Triple Integral",
    journal="The American Mathematical Monthly",
    volume="47",
    number="8",
    month="oct",
    year="1940",
    pages="539-543",
    url="https://doi.org/10.1080/00029890.1940.11991016",
)


def sadowsky():
    d = {
        "symm_r00": [[frac(91, 450)], [1]],
        "symm_rr0": [[frac(-20, 225)], [1]],
        "symm_rrs": [[frac(8, 225)], [sqrt(frac(5, 8))], [1]],
    }
    return C3Scheme("Sadowsky", d, 5, source)


register([sadowsky])



import math

import numpy as np
import sympy

from ..helpers import article, get_nsimplex_points, untangle, z
from ._helpers import SnScheme

source = article(
    authors=["Srebra B. Stoyanova"],
    title="Cubature formulae of the seventh degree of accuracy for the hypersphere",
    journal="Journal of Computational and Applied Mathematics",
    year="1997",
    pages="15-21",
    url="https://doi.org/10.1016/S0377-0427%2897%2900094-0",
)


def stoyanova(n, delta=None, variant_v_plus=True, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    assert n >= 5

    if delta is None:
        # pick a delta that "works" (points real-valued etc.)
        if variant_v_plus:
            if n < 10:
                delta = {5: 0.98945, 6: 1.005, 7: 1.017, 8: 1.026, 9: 1.04}[n]
            else:
                delta = 1
        else:
            if n < 10:
                delta = {5: 0.98928, 6: 0.818, 7: 0.81, 8: 0.8, 9: 0.8}[n]
            else:
                delta = sympy.Rational(4, 5)

    pts_a = get_nsimplex_points(n, sqrt, frac)

    # simplex edge midpoints projected onto the sphere
    pts_b = np.array(
        [
            sqrt(frac(n, 2 * (n - 1))) * (pts_a[i1] + pts_a[i2])
            for i1 in range(n + 1)
            for i2 in range(i1)
        ]
    )
    # face midpoints projected onto the sphere
    pts_c = np.array(
        [
            sqrt(frac(n, 3 * (n - 2))) * (pts_a[i1] + pts_a[i2] + pts_a[i3])
            for i1 in range(n + 1)
            for i2 in range(i1)
            for i3 in range(i2)
        ]
    )
    # (1/4)-points on the edges connecting the simplex points
    pts_b14 = np.array(
        [
            sqrt(frac(n, 10 * n - 6)) * (pts_a[k] + 3 * pts_a[l])
            for k in range(n + 1)
            for l in range(k)
        ]
        + [
            sqrt(frac(n, 10 * n - 6)) * (3 * pts_a[k] + pts_a[l])
            for k in range(n + 1)
            for l in range(k)
        ]
    )

    pts_a = np.concatenate([pts_a, -pts_a])
    pts_b = np.concatenate([pts_b, -pts_b])
    if n > 5:
        # for n==5, the points are already symmetric
        pts_c = np.concatenate([pts_c, -pts_c])
    pts_b14 = np.concatenate([pts_b14, -pts_b14])

    n1 = 2 * (n + 1)
    n2 = n * (n + 1)
    n3 = (n + 1) * n * (n - 1) // 6 if n == 5 else (n + 1) * n * (n - 1) // 3
    n4 = 2 * n * (n + 1)

    assert len(pts_a) == n1
    assert len(pts_b) == n2
    assert len(pts_c) == n3
    assert len(pts_b14) == n4

    s1 = (n - 2) ** 5 + 243
    s2 = (n - 1) ** 5 + 32
    s3 = (3 * n - 1) ** 6 + (n - 3) ** 6 + 4096 * (n - 1)
    s4 = n ** 5 + 1
    s5 = n ** 2 + 4 * n - 8
    s6 = (n + 1) * (n + 2) * (n + 4)
    s7 = n * (n - 1)
    s8 = n * (n - 2)
    s9 = (n - 1) * (n - 2)
    s10 = (n + 1) * (n + 6) * s6
    s11 = (5 * n - 3) ** 3
    s12 = (n - 1) * (n - 3)
    s13 = n ** 2 - 7 * n + 19
    s14 = n ** 2 - n + 1
    s15 = 2 * (n - 3) * (n + 1) ** 2
    s16 = n ** 3 - 9 * n ** 2 + 33 * n - 38
    s17 = 3 * (n - 2)
    q1 = (
        9
        * (n + 1)
        * (
            10 * n ** 5 * (n + 1) ** 2 * s17
            - n ** 3 * s2 * s5
            + 4 * (n - 1) * s4 * s5 * s7
            - 2 * s4 * s6 * s8
        )
        - n ** 2 * s3 * s8
        + 8 * s4 * s8 * s11
        + 108 * n * s2 * s7 ** 2
        - 432 * (n - 1) ** 3 * s4 * s7
    )
    q2 = (
        n * s1 * s7
        - 3 * s4 * s9 * s17
        + 12 * (n - 1) * s4 * s12
        - 3 * n ** 2 * (n - 3) * s2
    )
    b1 = frac(
        2
        * (n - 1)
        * (3 * (n + 1) * s5 * s7 * q2 - s12 * q1 - 36 * (n - 1) ** 2 * s7 * q2),
        s10 * s17 * q2,
    )
    c1 = frac(s9 * q1, 2 * s10 * q2)
    e1 = frac(4 * n * s11, 9 * s10)
    a1 = frac(n, n + 6) - b1 - c1 - e1
    y1 = frac(n, n + 2) - frac(e1, delta ** 4)
    y2 = frac(n, n + 4) - frac(e1, delta ** 2)
    y3 = frac(3 * n ** 2, (n + 2) * (n + 4)) - frac(
        e1 * (41 * n ** 3 - 101 * n ** 2 + 155 * n - 87),
        2 * (5 * n - 3) ** 2 * delta ** 2,
    )
    w0 = frac(s17 * (y3 * (s15 - s14 * s17) + s13 * s14 * y2), s13 * s15 * c1)
    u0 = frac(n * (s17 * y3 - s13 * y2), s15 * a1)
    p1 = frac(n * (n - 4) * b1, 4 * s12 * a1)
    p2 = frac(s16 * s17 * b1, 4 * s12 * s13 * c1)
    d1 = a1 * p1 ** 2 + c1 * p2 ** 2 + b1
    d2 = a1 * p1 * u0 + c1 * p2 * w0
    d3 = a1 * u0 ** 2 + c1 * w0 ** 2 - y1
    d0 = d2 ** 2 - d1 * d3
    if variant_v_plus:
        v = (d2 + sqrt(d0)) / d1
    else:
        v = (d2 - sqrt(d0)) / d1
    u = u0 - p1 * v
    w = w0 - p2 * v
    lmbda2 = 1 / u
    beta2 = 1 / v
    gamma2 = 1 / w
    delta2 = delta ** 2

    lmbda6 = lmbda2 ** 3
    beta6 = beta2 ** 3
    gamma6 = gamma2 ** 3
    delta6 = delta2 ** 3

    a = a1 / n1 / lmbda6
    b = b1 / n2 / beta6
    c = c1 / n3 / gamma6
    e = e1 / n4 / delta6
    d = 1 - n1 * a - n2 * b - n3 * c - n4 * e

    lmbda = sqrt(lmbda2)
    beta = sqrt(beta2)
    gamma = sqrt(gamma2)

    data = [
        (a, lmbda * pts_a),
        (b, beta * pts_b),
        (c, gamma * pts_c),
        (e, delta * pts_b14),
        (d, z(n)),
    ]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return SnScheme("Stoyanova", n, weights, points, 7, source, 1.423e-14)



import numpy as np
import sympy

from ..helpers import book
from ._helpers import U2Scheme, register

# Pages 73-74 in
_source = book(
    authors="V.I. Krylov",
    title="Approximate Calculation of Integrals",
    publisher="Macmillan, New York",
    year="1962",
    note="Translated from 1st Russian ed., 1959, by A.H. Stroud",
    url="https://books.google.de/books/about/Approximate_Calculation_of_Integrals.html?id=ELeRwR27IRIC",
)

frac = sympy.Rational
cos = np.vectorize(sympy.cos)
sin = np.vectorize(sympy.sin)
pi = sympy.pi


def krylov(n: int):
    weights = np.full(n, frac(1, n))
    alpha = 2 * np.arange(n) * pi / n
    points = np.column_stack([cos(alpha), sin(alpha)])
    return U2Scheme(f"Krylov {n}", _source, n - 1, weights, points, 2.145e-16)


register([krylov])



# ENH closed forms of all schemes
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import C2Scheme, register

source = article(
    authors=["Preston C. Hammer", "Arthur H. Stroud"],
    title="Numerical Evaluation of Multiple Integrals II",
    journal="Math. Comp.",
    volume="12",
    year="1958",
    pages="272-280",
    url="https://doi.org/10.1090/S0025-5718-1958-0102176-6",
)


def hammer_stroud_1_2():
    d = {"d4_a0": [[frac(1, 4)], [sqrt(frac(2, 3))]]}
    return C2Scheme("Hammer-Stroud 1-2", d, 3, source)


def hammer_stroud_2_2():
    alpha = sqrt(frac(3, 5))
    d = {
        "zero2": [[frac(16, 81)]],
        "d4_a0": [[frac(10, 81)], [alpha]],
        "d4_aa": [[frac(25, 4 * 81)], [alpha]],
    }
    return C2Scheme("Hammer-Stroud 2-2", d, 5, source)


def hammer_stroud_3_2():
    xi1, xi2 = (sqrt(frac(3, 287) * (38 - i * sqrt(583))) for i in [+1, -1])
    d = {
        "d4_a0": [[frac(98, 4 * 405)], [sqrt(frac(6, 7))]],
        "d4_aa": [[0.5205929166673945 / 4, 0.2374317746906302 / 4], [xi1, xi2]],
    }
    return C2Scheme("Hammer-Stroud 3-2", d, 7, source, 4.441e-16)


register([hammer_stroud_1_2, hammer_stroud_2_2, hammer_stroud_3_2])



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import E2r2Scheme, register

_source = article(
    authors=["A.H. Stroud", "D. Secrest"],
    title="Approximate integration formulas for certain spherically symmetric regions",
    journal="Math. Comp.",
    volume="17",
    year="1963",
    pages="105-135",
    url="https://doi.org/10.1090/S0025-5718-1963-0161473-0",
)


def stroud_secrest_5():
    nu = sqrt(2)
    xi = nu / 2
    eta = sqrt(6) / 2
    A = frac(1, 2)
    B = frac(1, 12)

    d = {"zero2": [[A]], "c2_a0": [[B], [nu]], "sxy": [[B], [xi], [eta]]}
    return E2r2Scheme("Stroud-Secrest V", d, 5, _source, 4.360e-16)


def stroud_secrest_6():
    sqrt5 = sqrt(5)
    nu = sqrt(3)
    xi, eta = (sqrt((9 - p_m * 3 * sqrt5) / 8) for p_m in [+1, -1])
    A = frac(1, 36)
    B, C = ((5 + p_m * 2 * sqrt5) / 45 for p_m in [+1, -1])

    d = {"d4_a0": [[A], [nu]], "d4_aa": [[B, C], [xi, eta]]}
    return E2r2Scheme("Stroud-Secrest VI", d, 7, _source, 6.661e-16)


register([stroud_secrest_5, stroud_secrest_6])



import math
from collections import namedtuple

import numpy as np
import orthopy

from ..helpers import article
from ..tools import scheme_from_rc
from ._gauss_legendre import gauss_legendre
from ._helpers import C1Scheme, _find_shapes

source = article(
    authors=["Dirk P. Laurie"],
    title="Calculation of Gauss-Kronrod quadrature rules",
    journal="Math. Comp.",
    volume="66",
    year="1997",
    pages="1133-1145",
    url="https://doi.org/10.1090/S0025-5718-97-00861-2",
)


def gauss_kronrod(n, a=0, b=0):
    """
    Gauss-Kronrod quadrature; see
    <https://en.wikipedia.org/wiki/Gauss%E2%80%93Kronrod_quadrature_formula>.

    Besides points and weights, this class provides the weights of the corresponding
    Gauss-Legendre scheme in self.gauss_weights.

    Code adapted from
    <https://www.cs.purdue.edu/archives/2002/wxg/codes/r_kronrod.m>,
    <https://www.cs.purdue.edu/archives/2002/wxg/codes/kronrod.m>.
    """
    # The general scheme is:
    # Get the Jacobi recurrence coefficients, get the Kronrod vectors alpha and beta,
    # and hand those off to scheme_from_rc. There, the eigenproblem for a tridiagonal
    # matrix with alpha and beta is solved to retrieve the points and weights.
    # TODO replace math.ceil by -(-k//n)
    length = int(math.ceil(3 * n / 2.0)) + 1
    degree = 2 * length + 1

    rc = orthopy.c1.jacobi.RecurrenceCoefficients("monic", a, b, symbolic=False)
    _, alpha, beta = np.array([rc[k] for k in range(length)]).T
    beta[0] = rc.int_1

    a, b = _r_kronrod(n, alpha, beta)
    x, w = scheme_from_rc(a, b, rc.int_1, mode="numpy")
    # sort by x
    i = np.argsort(x)
    points = x[i]
    weights = w[i]
    return C1Scheme(f"Gauss-Kronrod ({n})", degree, weights, points, source)


def _r_kronrod(n, a0, b0):
    assert len(a0) == int(math.ceil(3 * n / 2.0)) + 1
    assert len(b0) == int(math.ceil(3 * n / 2.0)) + 1

    a = np.zeros(2 * n + 1)
    b = np.zeros(2 * n + 1)

    k = int(math.floor(3 * n / 2.0)) + 1
    a[:k] = a0[:k]
    k = int(math.ceil(3 * n / 2.0)) + 1
    b[:k] = b0[:k]
    s = np.zeros(int(math.floor(n / 2.0)) + 2)
    t = np.zeros(int(math.floor(n / 2.0)) + 2)
    t[1] = b[n + 1]
    for m in range(n - 1):
        k0 = int(math.floor((m + 1) / 2.0))
        k = np.arange(k0, -1, -1)
        L = m - k
        s[k + 1] = np.cumsum(
            (a[k + n + 1] - a[L]) * t[k + 1] + b[k + n + 1] * s[k] - b[L] * s[k + 1]
        )
        s, t = t, s

    j = int(math.floor(n / 2.0)) + 1
    s[1 : j + 1] = s[:j]
    for m in range(n - 1, 2 * n - 2):
        k0 = m + 1 - n
        k1 = int(math.floor((m - 1) / 2.0))
        k = np.arange(k0, k1 + 1)
        L = m - k
        j = n - 1 - L
        s[j + 1] = np.cumsum(
            -(a[k + n + 1] - a[L]) * t[j + 1]
            - b[k + n + 1] * s[j + 1]
            + b[L] * s[j + 2]
        )
        j = j[-1]
        k = int(math.floor((m + 1) / 2.0))
        if m % 2 == 0:
            a[k + n + 1] = a[k] + (s[j + 1] - b[k + n + 1] * s[j + 2]) / t[j + 2]
        else:
            b[k + n + 1] = s[j + 1] / s[j + 2]
        s, t = t, s

    a[2 * n] = a[n - 1] - b[2 * n] * s[1] / t[1]
    return a, b


def _gauss_kronrod_integrate(
    k,
    f,
    intervals,
    dot=np.dot,
    domain_shape=None,
    range_shape=None,
):
    # Compute the integral estimations according to Gauss and Gauss-Kronrod, sharing the
    # function evaluations
    gk = gauss_kronrod(k)
    gl = gauss_legendre(k)
    # scale points
    x0 = 0.5 * (1.0 - gk.points)
    x1 = 0.5 * (1.0 + gk.points)

    sp = np.multiply.outer(intervals[0], x0) + np.multiply.outer(intervals[1], x1)
    if domain_shape is not None and range_shape is not None:
        # Make it easy on the f by flattening out interval_set
        interval_set_shape = intervals.shape[1 + len(domain_shape) :]
        sp = sp.reshape(*domain_shape, -1)
        fx_gk = np.asarray(f(sp))
        fx_gk = fx_gk.reshape(*range_shape, *interval_set_shape, *x0.shape)
    else:
        fx_gk = np.asarray(f(sp))
        # try and guess shapes of domain, range, intervals
        domain_shape, range_shape, interval_set_shape = _find_shapes(
            fx_gk, intervals, gk.points, domain_shape, range_shape
        )

    fx_gl = fx_gk[..., 1::2]

    diff = intervals[1] - intervals[0]
    interval_lengths = np.sqrt(np.sum(diff ** 2, axis=tuple(range(len(domain_shape)))))
    assert interval_lengths.shape == interval_set_shape

    # integrate
    # average value of gk over the interval
    average_gk = 0.5 * dot(fx_gk, gk.weights)
    val_gauss_kronrod = average_gk * interval_lengths
    val_gauss_legendr = 0.5 * interval_lengths * dot(fx_gl, gl.weights)

    assert val_gauss_kronrod.shape == range_shape + interval_set_shape
    assert val_gauss_legendr.shape == range_shape + interval_set_shape

    # Get an error estimate. According to
    #
    #   Pedro Gonnet,
    #   A Review of Error Estimation in Adaptive Quadrature,
    #   ACM Computing Surveys (CSUR) Surveys,
    #   Volume 44, Issue 4, August 2012
    #   <https://doi.org/10.1145/2333112.2333117>,
    #   <https://arxiv.org/pdf/1003.4629.pdf>
    #
    # the classical QUADPACK still compares favorably with other approaches.
    fx_avg_abs = np.abs(fx_gk - average_gk[..., None])
    I_tilde = 0.5 * interval_lengths * dot(fx_avg_abs, gk.weights)

    # The exponent 1.5 is chosen such that (200*x)**1.5 is approximately x at 1.0e-6,
    # the machine precision on IEEE 754 32-bit floating point arithmentic. This could be
    # adapted to
    #
    #   eps = np.finfo(float).eps
    #   exponent = np.log(eps) / np.log(200*eps)
    #
    # The following expression is
    #
    # error_estimate = I_tilde * np.minimum(
    #     np.ones(I_tilde.shape),
    #     (200 * abs(val_gauss_kronrod - val_gauss_legendr) / I_tilde) ** 1.5,
    # )
    #
    # with handling NaNs (if I_tilde is 0).
    error_estimate = np.empty(I_tilde.shape)
    idx = np.abs(I_tilde) > 1.0e-15
    vals = (
        200 * np.abs(val_gauss_kronrod[idx] - val_gauss_legendr[idx]) / I_tilde[idx]
    ) ** 1.5
    error_estimate[idx] = np.minimum(np.ones(np.sum(idx)), vals)
    error_estimate[~idx] = 1.0
    error_estimate *= I_tilde

    assert error_estimate.shape == range_shape + interval_set_shape

    gk_return = namedtuple(
        "gauss_kronrod_return_values",
        [
            "val_gauss_kronrod",
            "val_gauss_legendre",
            "interval_lengths",
            "error_estimate",
            "domain_shape",
            "range_shape",
        ],
    )

    return gk_return(
        val_gauss_kronrod,
        val_gauss_legendr,
        interval_lengths,
        error_estimate,
        domain_shape,
        range_shape,
    )



import numpy as np

from ._helpers import C3Scheme


def product(scheme1d):
    schemes = scheme1d if isinstance(scheme1d, list) else 3 * [scheme1d]

    wy, wz, wx = np.meshgrid(schemes[0].weights, schemes[1].weights, schemes[2].weights)
    weights = np.vstack([wx.flatten(), wy.flatten(), wz.flatten()]).T
    weights = np.prod(weights, axis=1)
    assert len(weights) > 0
    # the order, yeah...
    y, z, x = np.meshgrid(schemes[0].points, schemes[1].points, schemes[2].points)
    points = np.array([x.flatten(), y.flatten(), z.flatten()])

    degree = min(s.degree for s in schemes)
    weights /= 8

    d = {"plain": [weights, points[0], points[1], points[2]]}
    return C3Scheme(f"Product scheme ({scheme1d.name})", d, degree)



from . import (
    _beckers_haegemans,
    _gatermann,
    _hammer_marlowe_stroud,
    _hammer_stroud,
    _jaskowiec_sukumar,
    _keast,
    _liu_vinokur,
    _maeztu_sainz,
    _shunn_ham,
    _stroud,
    _vioreanu_rokhlin,
    _walkington,
    _williams_shunn_jameson,
    _witherden_vincent,
    _xiao_gimbutas,
    _yu,
    _zhang_cui_liu,
)
from ._get_good_scheme import get_good_scheme
from ._helpers import schemes

__all__ = [
    "_beckers_haegemans",
    "_gatermann",
    "_hammer_marlowe_stroud",
    "_hammer_stroud",
    "_jaskowiec_sukumar",
    "_keast",
    "_liu_vinokur",
    "_maeztu_sainz",
    "_shunn_ham",
    "_stroud",
    "_vioreanu_rokhlin",
    "_walkington",
    "_williams_shunn_jameson",
    "_witherden_vincent",
    "_xiao_gimbutas",
    "_yu",
    "_zhang_cui_liu",
    #
    "schemes",
    "get_good_scheme",
]



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, comb, fsd, untangle, z
from ._helpers import CnScheme

_source = article(
    authors=["G.M. Phillips"],
    title="Numerical integration over an N-dimensional rectangular region",
    journal="Comput J",
    year="1967",
    volume="10",
    number="3",
    pages="297-299",
    url="https://doi.org/10.1093/comjnl/10.3.297",
)


def phillips(n):
    if n == 2:
        p1 = 1
        p2 = frac(14, 3)
        q = frac(5, 3)
    elif n == 3:
        p1 = 1
        p2 = frac(14, 5)
        q = frac(5, 2)
        r = 1
    elif n == 4:
        p1 = 1
        p2 = frac(112, 11)
        q = 5
        r = 2
    else:
        assert n >= 5
        p1 = 1
        En = frac(25 * n ** 2 - 165 * n + 302, 972)
        p2 = 1 / (frac(3, 5) - frac(1, 35 * En))
        q = frac(5, 3)
        r = frac(5, 3)

    gamma = frac((n - 1) * (19 - 5 * n), 270)
    delta = frac((n - 1) * (n - 2), 108)

    a1 = frac(23 - 5 * n, 180) - gamma * q / 2
    a2 = frac(35 * n ** 2 - 231 * n + 466, 3780)
    beta1 = (a1 - a2 * p2) / (p1 - p2)
    beta2 = (a1 - a2 * p1) / (p2 - p1)

    lambda1 = 1 / sqrt(p1)
    lambda2 = 1 / sqrt(p2)
    mu = 1 / sqrt(q)

    b1 = beta1 / lambda1 ** 6
    b2 = beta2 / lambda2 ** 6

    c = gamma / (2 * (n - 1) * mu ** 6)

    a = 1 - 2 * n * (b1 + b2) - 4 * comb(n, 2) * c

    if n > 2:
        nu = 1 / sqrt(r)
        d = delta / (4 * comb(n - 1, 2) * nu ** 6)
        a -= 8 * comb(n, 3) * d

    data = [
        (a, z(n)),
        (b1, fsd(n, (lambda1, 1))),
        (b2, fsd(n, (lambda2, 1))),
        (c, fsd(n, (mu, 2))),
    ]

    if n > 2:
        data.append((d, fsd(n, (nu, 3))))

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Phillips", n, weights, points, 7, _source, 1.521e-13)



import numpy as np

from ..tn import get_vol
from ._dunavant import dunavant_05, dunavant_10


def _numpy_all_except(a, axis=-1):
    axes = np.arange(a.ndim)
    axes = np.delete(axes, axis)
    return np.all(a, axis=tuple(axes))


def integrate_adaptive(
    f,
    triangles,
    eps,
    minimum_triangle_area=None,
    scheme1=dunavant_05(),
    scheme2=dunavant_10(),
    dot=np.dot,
):
    sumfun = np.sum

    triangles = np.array(triangles)
    if len(triangles.shape) == 2:
        # add dimension in the second-to-last place
        triangles = np.expand_dims(triangles, -2)

    areas = get_vol(triangles)
    total_area = sumfun(areas)

    if minimum_triangle_area is None:
        minimum_triangle_area = total_area * 0.25 ** 10

    val1 = scheme1.integrate(f, triangles, dot=dot)
    val2 = scheme2.integrate(f, triangles, dot=dot)
    error_estimate = abs(val1 - val2)

    # Mark intervals with acceptable approximations. For this, take all()
    # across every dimension except the last one, which is the interval index.
    is_good = _numpy_all_except(error_estimate < eps * areas / total_area, axis=-1)

    # add values from good intervals to sum
    quad_sum = sumfun(val1[..., is_good], axis=-1)
    global_error_estimate = sumfun(error_estimate[..., is_good], axis=-1)

    is_bad = np.logical_not(is_good)
    while any(is_bad):
        # split the bad triangles into four #triforce
        #
        #         /\
        #        /__\
        #       /\  /\
        #      /__\/__\
        #
        triangles = triangles[..., is_bad, :]
        midpoints = [
            0.5 * (triangles[1] + triangles[2]),
            0.5 * (triangles[2] + triangles[0]),
            0.5 * (triangles[0] + triangles[1]),
        ]
        triangles = np.array(
            [
                np.concatenate(
                    [triangles[0], triangles[1], triangles[2], midpoints[0]]
                ),
                np.concatenate(
                    [midpoints[2], midpoints[0], midpoints[1], midpoints[1]]
                ),
                np.concatenate(
                    [midpoints[1], midpoints[2], midpoints[0], midpoints[2]]
                ),
            ]
        )
        areas = get_vol(triangles)
        assert all(areas > minimum_triangle_area)

        # compute values and error estimates for the new intervals
        val1 = scheme1.integrate(f, triangles, dot=dot)
        val2 = scheme2.integrate(f, triangles, dot=dot)
        error_estimate = abs(val1 - val2)

        # mark good intervals, gather values and error estimates
        is_good = _numpy_all_except(error_estimate < eps * areas / total_area, axis=-1)
        # add values from good intervals to sum
        quad_sum += sumfun(val1[..., is_good], axis=-1)
        global_error_estimate += sumfun(error_estimate[..., is_good], axis=-1)
        is_bad = np.logical_not(is_good)

    return quad_sum, global_error_estimate



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["D.P. Laurie"],
    title="Algorithm 584: CUBTRI: Automatic Cubature over a Triangle",
    journal="ACM Trans. Math. Softw.",
    month="jun",
    year="1982",
    url="https://doi.org/10.1145/355993.356001",
)

this_dir = pathlib.Path(__file__).resolve().parent


def cubtri():
    """
    Se also
    https://people.sc.fsu.edu/~jburkardt/datasets/quadrature_rules_tri/quadrature_rules_tri.html
    """
    return _read(this_dir / "cubtri.json", source)


register([cubtri])



import math

import numpy as np
import sympy

from ._helpers import C1Scheme


def newton_cotes_closed(index, **kwargs):
    """
    Closed Newton-Cotes formulae.
    <https://en.wikipedia.org/wiki/Newton%E2%80%93Cotes_formulas#Closed_Newton.E2.80.93Cotes_formulae>,
    <http://mathworld.wolfram.com/Newton-CotesFormulas.html>.
    """
    points = np.linspace(-1.0, 1.0, index + 1)
    degree = index + 1 if index % 2 == 0 else index

    # Formula (26) from
    # <http://mathworld.wolfram.com/Newton-CotesFormulas.html>.
    # Note that Sympy carries out all operations in rationals, i.e.,
    # _exactly_. Only at the end, the rational is converted into a float.
    n = index
    weights = np.empty(n + 1)
    t = sympy.Symbol("t")
    for r in range(n + 1):
        # Compare with get_weights().
        f = sympy.prod([(t - i) for i in range(n + 1) if i != r])
        alpha = (
            2
            * (-1) ** (n - r)
            * sympy.integrate(f, (t, 0, n), **kwargs)
            / (math.factorial(r) * math.factorial(n - r))
            / index
        )
        weights[r] = alpha
    return C1Scheme("Newton-Cotes (closed)", degree, weights, points)


def newton_cotes_open(index, **kwargs):
    """
    Open Newton-Cotes formulae.
    <https://math.stackexchange.com/a/1959071/36678>
    """
    points = np.linspace(-1.0, 1.0, index + 2)[1:-1]
    degree = index if (index + 1) % 2 == 0 else index - 1
    #
    n = index + 1
    weights = np.empty(n - 1)
    t = sympy.Symbol("t")
    for r in range(1, n):
        # Compare with get_weights().
        f = sympy.prod([(t - i) for i in range(1, n) if i != r])
        alpha = (
            2
            * (-1) ** (n - r + 1)
            * sympy.integrate(f, (t, 0, n), **kwargs)
            / (math.factorial(r - 1) * math.factorial(n - 1 - r))
            / n
        )
        weights[r - 1] = alpha
    return C1Scheme("Newton-Cotes (open)", degree, weights, points)



import math

from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import C2Scheme, register

source = article(
    authors=["A.M. Cohen", "D.A. Gismalla"],
    title="Some integration formulae for symmetric functions of two variables",
    journal="International Journal of Computer Mathematics",
    year="1986",
    volume="19",
    number="1",
    pages="57-68",
    url="https://doi.org/10.1080/00207168608803504",
)


def cohen_gismalla_1():
    B = frac(5, 28)
    u, v = (sqrt((frac(1, 3) + i * sqrt(frac(2, 63))) / frac(5, 7)) for i in [+1, -1])
    d = {"zero2": [[frac(2, 7)]], "c2": [[B, B], [u, v], [-v, u]]}
    # This scheme is of order 5 for symmetric integrands
    return C2Scheme("Cohen-Gismalla 1", d, 3, source, 4.996e-16)


def cohen_gismalla_2():
    # TODO improve precision
    X = 0.3850907
    Y = (24 / 35 - 28 / 45 * X) / (28 / 45 - 2 / 3 * X)
    alpha = (2 / 3 * Y - 28 / 45) / (Y - X)
    beta = (28 / 45 - 2 / 3 * X) / (Y - X)

    B = alpha / X / 4
    C = beta / Y / 4
    A = 1 - 4 * B - 4 * C

    g1 = alpha / B / 2
    g2 = X * (Y / 9 - 2 / 15) / (Y - X) / alpha
    u = math.sqrt(g1 + math.sqrt(g1 ** 2 - g2))
    v = math.sqrt(g1 - math.sqrt(g1 ** 2 - g2))

    h1 = beta / C / 2
    h2 = Y * (2 / 15 - X / 9) / (Y - X) / beta
    r = math.sqrt(h1 - math.sqrt(h1 ** 2 - h2))
    s = math.sqrt(h1 + math.sqrt(h1 ** 2 - h2))

    d = {"zero2": [[A]], "c2": [[B, B, C, C], [u, v, r, r], [-v, u, -s, s]]}
    # ERR this scheme only has order 1
    # According to the article, it has order 7 for symmetric integrands.
    # Something is fishy...
    return C2Scheme("Cohen-Gismalla 2", d, 1, source, 4.441e-16)


register([cohen_gismalla_1, cohen_gismalla_2])



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, expand_symmetries
from ._helpers import CnScheme

_source = article(
    authors=["D. Mustard", "J.N. Lyness", "J.M. Blatt"],
    title="Numerical quadrature in n dimensions",
    journal="Comput J",
    year="1963",
    volume="6",
    number="1",
    pages="75-87",
    url="https://doi.org/10.1093/comjnl/6.1.75",
)


def mustard_lyness_blatt(n):
    r = sqrt(frac(2, 5))
    d = {
        "0": [[frac(8 - 5 * n, 9)]],
        "a0": [[frac(5, 18)], [r]],
        "a": [[frac(1, 9 * 2 ** n)], [1]],
    }
    points, weights = expand_symmetries(d, n)
    return CnScheme("Mustard-Lyness-Blatt", n, weights, points, 5, _source, 6.312e-14)



import pathlib

from ...helpers import article
from .._helpers import _read, register

# Sphere integration schemes from a series of publications, in chronological order
# <https://en.wikipedia.org/wiki/Lebedev_quadrature>
# <https://people.sc.fsu.edu/~jburkardt/datasets/sphere_lebedev_rule/sphere_lebedev_rule.html>
sources = [
    article(
        authors=["V.I. Lebedev"],
        title="Values of the nodes and weights of ninth to seventeenth order Gauss-Markov quadrature formulae invariant under the octahedron group with inversion",
        journal="Computational Mathematics and Mathematical Physics",
        volume="15",
        year="1975",
        pages="44-51",
        url="https://doi.org/10.1016/0041-5553%2875%2990133-0",
    ),
    article(
        authors=["V.I. Lebedev"],
        year="1976",
        title="Quadratures on a sphere",
        journal="Zh. Vȳchisl. Mat. Mat. Fiz.",
        volume="16",
        number="2",
        pages="293–306",
        url="https://doi.org/10.1016/0041-5553%2876%2990100-2",
    ),
    article(
        authors=["V.I. Lebedev"],
        title="Spherical quadrature formulas exact to orders 25-29",
        journal="Siberian Mathematical Journal",
        volume="18",
        year="1977",
        pages="99-107",
        url="https://doi.org/10.1007/BF00966954",
    ),
    article(
        authors=["V.I. Lebedev", "A.L. Skorokhodov"],
        title="Quadrature formulas of orders 41, 47, and 53 for the sphere",
        journal="Russian Acad. Sci. Dokl. Math.",
        volume="45",
        year="1992",
        pages="587-592",
    ),
    article(
        authors=["V.I. Lebedev"],
        title="A quadrature formula for the sphere of 59th algebraic order of accuracy",
        journal="Russian Acad. Sci. Dokl. Math.",
        volume="50",
        year="1995",
        pages="283-286",
    ),
    article(
        authors=["V.I. Lebedev", "D.N. Laikov"],
        title="A quadrature formula for the sphere of the 131st algebraic order of accuracy",
        journal="Doklady Mathematics",
        volume="59",
        number="3",
        year="1999",
        pages="477-481",
    ),
]

this_dir = pathlib.Path(__file__).resolve().parent


def lebedev_003a():
    return _read(this_dir / "lebedev_003a.json", sources[0])


def lebedev_003b():
    return _read(this_dir / "lebedev_003b.json", sources[0])


def lebedev_003c():
    return _read(this_dir / "lebedev_003c.json", sources[0])


def lebedev_005():
    return _read(this_dir / "lebedev_005.json", sources[0])


def lebedev_007():
    return _read(this_dir / "lebedev_007.json", sources[0])


def lebedev_009():
    return _read(this_dir / "lebedev_009.json", sources[0])


def lebedev_011():
    return _read(this_dir / "lebedev_011.json", sources[0])


def lebedev_013():
    return _read(this_dir / "lebedev_013.json", sources[0])


def lebedev_015():
    return _read(this_dir / "lebedev_015.json", sources[0])


def lebedev_017():
    return _read(this_dir / "lebedev_017.json", sources[0])


def lebedev_019():
    return _read(this_dir / "lebedev_019.json", sources[1])


def lebedev_021():
    return _read(this_dir / "lebedev_021.json", sources[1])


def lebedev_023():
    return _read(this_dir / "lebedev_023.json", sources[1])


def lebedev_025():
    return _read(this_dir / "lebedev_025.json", sources[2])


def lebedev_027():
    return _read(this_dir / "lebedev_027.json", sources[2])


def lebedev_029():
    return _read(this_dir / "lebedev_029.json", sources[2])


def lebedev_031():
    return _read(this_dir / "lebedev_031.json", sources[3])


def lebedev_035():
    return _read(this_dir / "lebedev_035.json", sources[3])


def lebedev_041():
    return _read(this_dir / "lebedev_041.json", sources[3])


def lebedev_047():
    return _read(this_dir / "lebedev_047.json", sources[3])


def lebedev_053():
    return _read(this_dir / "lebedev_053.json", sources[3])


def lebedev_059():
    return _read(this_dir / "lebedev_059.json", sources[4])


def lebedev_065():
    return _read(this_dir / "lebedev_065.json", sources[5])


def lebedev_071():
    return _read(this_dir / "lebedev_071.json", sources[5])


def lebedev_077():
    return _read(this_dir / "lebedev_077.json", sources[5])


def lebedev_083():
    return _read(this_dir / "lebedev_083.json", sources[5])


def lebedev_089():
    return _read(this_dir / "lebedev_089.json", sources[5])


def lebedev_095():
    return _read(this_dir / "lebedev_095.json", sources[5])


def lebedev_101():
    return _read(this_dir / "lebedev_101.json", sources[5])


def lebedev_107():
    return _read(this_dir / "lebedev_107.json", sources[5])


def lebedev_113():
    return _read(this_dir / "lebedev_113.json", sources[5])


def lebedev_119():
    return _read(this_dir / "lebedev_119.json", sources[5])


def lebedev_125():
    return _read(this_dir / "lebedev_125.json", sources[5])


def lebedev_131():
    return _read(this_dir / "lebedev_131.json", sources[5])


register(
    [
        lebedev_003a,
        lebedev_003b,
        lebedev_003c,
        lebedev_005,
        lebedev_007,
        lebedev_009,
        lebedev_011,
        lebedev_013,
        lebedev_015,
        lebedev_017,
        lebedev_019,
        lebedev_021,
        lebedev_023,
        lebedev_025,
        lebedev_027,
        lebedev_029,
        lebedev_031,
        lebedev_035,
        lebedev_041,
        lebedev_047,
        lebedev_053,
        lebedev_059,
        lebedev_065,
        lebedev_071,
        lebedev_077,
        lebedev_083,
        lebedev_089,
        lebedev_095,
        lebedev_101,
        lebedev_107,
        lebedev_113,
        lebedev_119,
        lebedev_125,
        lebedev_131,
    ]
)



import sympy


def volume(n, symbolic=False):
    n = sympy.S(n) if symbolic else n

    def _recurrence(n):
        if n == 0:
            return 1
        return _recurrence(n - 1) / n

    return _recurrence(n)


def integrate_monomial(exponents, symbolic=False):
    """The integrals of monomials over the standard triangle and tetrahedron are
    given by

    \\int_T x_0^k0 * x1^k1 = (k0!*k1!) / (2+k0+k1)!,
    \\int_T x_0^k0 * x1^k1 * x2^k2 = (k0!*k1!*k2!) / (3+k0+k1+k2)!,

    see, e.g.,
    A set of symmetric quadrature rules on triangles and tetrahedra,
    Linbo Zhang, Tao Cui and Hui Liu,
    Journal of Computational Mathematics,
    Vol. 27, No. 1 (January 2009), pp. 89-96,
    <https://www.jstor.org/stable/43693493>.

    See, e.g., <https://math.stackexchange.com/q/207073/36678> for a formula in
    all dimensions.

    To cope with the huge terms in numerator and denominator, it might make sense to use
    exp(lgamma()). It's even easier though to represent the above expression by the
    recurrence with the simple factor k_i / (sum(k) + n) which will never overflow. It's
    also well-suited for symbolic computation.
    """
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    exponents = list(exponents)
    n = len(exponents)

    def _recurrence(exponents):
        if all(k == 0 for k in exponents):
            return volume(n, symbolic)

        # find first nonzero
        idx, k0 = next((i, k) for i, k in enumerate(exponents) if k > 0)
        k2 = exponents.copy()
        k2[idx] -= 1
        return _recurrence(k2) * frac(k0, sum(exponents) + n)

    return _recurrence(exponents)



import pathlib

from sympy import Rational as frac

from ...helpers import article
from .._classical import centroid as laursen_gellert_01
from .._helpers import T2Scheme, _read, register
from .._strang_fix_cowper import strang_fix_cowper_03 as laursen_gellert_03  # DUP
from .._strang_fix_cowper import strang_fix_cowper_04 as laursen_gellert_04
from .._strang_fix_cowper import strang_fix_cowper_05 as laursen_gellert_05

source = article(
    authors=["M.E. Laursen", "M. Gellert"],
    title="Some criteria for numerically integrated matrices and quadrature formulas for triangles",
    journal="International Journal for Numerical Methods in Engineering",
    volume="12",
    number="1",
    year="1978",
    pages="67–76",
    url="https://doi.org/10.1002/nme.1620120107",
)

this_dir = pathlib.Path(__file__).resolve().parent


def laursen_gellert_02a():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 6)]]}
    return T2Scheme("Laursen-Gellert 2a", d, 2, source)


def laursen_gellert_02b():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 2)]]}
    return T2Scheme("Laursen-Gellert 2b", d, 2, source)


def laursen_gellert_06():
    d = {
        "centroid": [[3 / 8]],
        "d3_ab": [[5 / 48], [0.736712498968435], [0.237932366472434]],
    }
    return T2Scheme("Laursen-Gellert 6", d, 4, source)


def laursen_gellert_07():
    d = {
        "centroid": [[9 / 40]],
        "d3_aa": [
            [0.125939180544827, 0.132394152788506],
            [0.101286507323456, 0.470142064105115],
        ],
    }
    return T2Scheme("Laursen-Gellert 7", d, 5, source)


def laursen_gellert_08():
    return _read(this_dir / "laursen_gellert_08.json", source)


def laursen_gellert_09():
    return _read(this_dir / "laursen_gellert_09.json", source)


def laursen_gellert_10():
    return _read(this_dir / "laursen_gellert_10.json", source)


def laursen_gellert_11():
    return _read(this_dir / "laursen_gellert_11.json", source)


def laursen_gellert_12():
    return _read(this_dir / "laursen_gellert_12.json", source)


def laursen_gellert_13():
    return _read(this_dir / "laursen_gellert_13.json", source)


def laursen_gellert_14():
    return _read(this_dir / "laursen_gellert_14.json", source)


def laursen_gellert_15a():
    return _read(this_dir / "laursen_gellert_15a.json", source)


def laursen_gellert_15b():
    return _read(this_dir / "laursen_gellert_15b.json", source)


register(
    [
        laursen_gellert_01,
        laursen_gellert_02a,
        laursen_gellert_02b,
        laursen_gellert_03,
        laursen_gellert_04,
        laursen_gellert_05,
        laursen_gellert_06,
        laursen_gellert_07,
        laursen_gellert_08,
        laursen_gellert_09,
        laursen_gellert_10,
        laursen_gellert_11,
        laursen_gellert_12,
        laursen_gellert_13,
        laursen_gellert_14,
        laursen_gellert_15a,
        laursen_gellert_15b,
    ]
)



import math

import numpy as np
import sympy

from ..helpers import article, comb, fsd, prod, untangle, z
from ._helpers import CnScheme

_source = article(
    authors=["J. McNamee", "F. Stenger"],
    title="Construction of Fully Symmetric Numerical Integration Formulas",
    journal="Numerische Mathematik",
    year="1967",
    volume="10",
    pages="327-344",
    url="https://doi.org/10.1007/BF02162032",
)


def _mcnamee_stenger_3(n, integrator, symbolic):
    sqrt = sympy.sqrt if symbolic else math.sqrt

    I0 = integrator(n, [], symbolic)
    I2 = integrator(n, [2], symbolic)

    u2 = I2 / I0
    # ERR The article says I0 / (2 * u**2)
    A1 = I2 / (2 * u2)
    A0 = (1 - n) * I0
    u = sqrt(u2)

    data = [
        (A0 / I0, z(n)),
        (A1 / I0, fsd(n, (u, 1))),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return "McNamee-Stenger 3", n, weights, points, 3, _source


def _mcnamee_stenger_5(n, integrator, symbolic):
    assert n >= 2
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt

    I0 = integrator(n, [], symbolic)
    I2 = integrator(n, [2], symbolic)
    I4 = integrator(n, [4], symbolic)
    I22 = integrator(n, [2, 2], symbolic)

    u = sqrt(I4 / I2)
    A0 = I0 - n * (I2 / I4) ** 2 * (I4 - frac(n - 1, 2) * I22)
    A1 = frac(1, 2) * (I2 / I4) ** 2 * (I4 - (n - 1) * I22)
    A11 = frac(1, 4) * (I2 / I4) ** 2 * I22

    data = [
        (A0, z(n)),
        (A1, fsd(n, (u, 1))),
        (A11, fsd(n, (u, 2))),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    weights /= I0
    return "McNamee-Stenger 5", n, weights, points, 5, _source


def _mcnamee_stenger_7(n, integrator, switch_uv, symbolic):
    assert n >= 3
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt

    I0 = integrator(n, [], symbolic)
    I2 = integrator(n, [2], symbolic)
    I4 = integrator(n, [4], symbolic)
    I6 = integrator(n, [6], symbolic)
    I22 = integrator(n, [2, 2], symbolic)
    I24 = integrator(n, [2, 4], symbolic)
    I222 = integrator(n, [2, 2, 2], symbolic)

    # Choose u, v as solutions of a u**4 - b u**2 + c = 0.
    a = I2 ** 2 - I0 * I4
    b = I2 * I4 - I0 * I6
    c = I4 ** 2 - I2 * I6
    #
    u2 = (b + sqrt(b ** 2 - 4 * a * c)) / (2 * a)
    v2 = (b - sqrt(b ** 2 - 4 * a * c)) / (2 * a)

    if switch_uv:
        u2, v2 = v2, u2

    u4 = u2 ** 2
    u6 = u2 ** 3
    v4 = v2 ** 2
    v6 = v2 ** 3

    A111 = I222 / (2 * u2) ** 3

    # mat = [[u4, v4], [u6, v6]]
    det = u4 * v6 - v4 * u6
    inv = [[v6 / det, -v4 / det], [-u6 / det, u4 / det]]
    # vector
    vec = [
        I22 - 2 ** 3 * (n - 2) * u4 * A111,
        I24 - 2 ** 3 * (n - 2) * u6 * A111,
    ]
    A11 = frac(1, 2 ** 2) * (inv[0][0] * vec[0] + inv[0][1] * vec[1])
    A22 = frac(1, 2 ** 2) * (inv[1][0] * vec[0] + inv[1][1] * vec[1])

    # mat = [[u2, v2], [u4, v4]]
    det = u2 * v4 - v2 * u4
    inv = [[v4 / det, -v2 / det], [-u4 / det, u2 / det]]
    vec = [
        I2 - 2 ** 3 * comb(n - 1, 2) * u2 * A111,
        I4 - 2 ** 3 * comb(n - 1, 2) * u4 * A111,
    ]
    A1 = -2 * (n - 1) * A11 + frac(1, 2) * (inv[0][0] * vec[0] + inv[0][1] * vec[1])
    A2 = -2 * (n - 1) * A22 + frac(1, 2) * (inv[1][0] * vec[0] + inv[1][1] * vec[1])

    A0 = (
        I0
        - 2 * n * (A1 + A2)
        - 2 ** 2 * comb(n, 2) * (A11 + A22)
        - 2 ** 3 * comb(n, 3) * A111
    )

    u = sqrt(u2)
    v = sqrt(v2)

    data = [
        (A0, z(n)),
        (A1, fsd(n, (u, 1))),
        (A2, fsd(n, (v, 1))),
        (A11, fsd(n, (u, 2))),
        (A22, fsd(n, (v, 2))),
        (A111, fsd(n, (u, 3))),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)

    weights /= I0
    variant = "a" if not switch_uv else "b"
    return f"McNamee-Stenger 7{variant}", n, weights, points, 7, _source


def _mcnamee_stenger_9(n, integrator, switch_uv, symbolic):
    assert n >= 4
    sqrt = sympy.sqrt if symbolic else math.sqrt
    frac = sympy.Rational if symbolic else lambda a, b: a / b

    I0 = integrator(n, [], symbolic)
    I2 = integrator(n, [2], symbolic)
    I4 = integrator(n, [4], symbolic)
    I6 = integrator(n, [6], symbolic)
    I8 = integrator(n, [8], symbolic)
    I24 = integrator(n, [2, 4], symbolic)
    I26 = integrator(n, [2, 6], symbolic)
    I44 = integrator(n, [4, 4], symbolic)
    I222 = integrator(n, [2, 2, 2], symbolic)
    I224 = integrator(n, [2, 2, 4], symbolic)
    I2222 = integrator(n, [2, 2, 2, 2], symbolic)

    # Choose u, v as solutions of a u**4 - b u**2 + c = 0.
    a = I4 ** 2 - I2 * I6
    b = I4 * I6 - I2 * I8
    c = I6 ** 2 - I4 * I8
    #
    u2 = (b + sqrt(b ** 2 - 4 * a * c)) / (2 * a)
    v2 = (b - sqrt(b ** 2 - 4 * a * c)) / (2 * a)

    if switch_uv:
        u2, v2 = v2, u2

    u4 = u2 ** 2
    u6 = u2 ** 3
    u8 = u2 ** 4
    v4 = v2 ** 2
    v6 = v2 ** 3
    v8 = v2 ** 4

    A1111 = I2222 / (2 * u2) ** 4

    # mat = [[u6, v6], [u8, v8]]
    det = u6 * v8 - v6 * u8
    inv = [[v8 / det, -v6 / det], [-u8 / det, u6 / det]]
    # vector
    vec = [
        I222 - 2 ** 4 * (n - 3) * u6 * A1111,
        I224 - 2 ** 4 * (n - 3) * u8 * A1111,
    ]
    A111 = frac(1, 2 ** 3) * (inv[0][0] * vec[0] + inv[0][1] * vec[1])
    A222 = frac(1, 2 ** 3) * (inv[1][0] * vec[0] + inv[1][1] * vec[1])

    A12 = (I26 - I44) / (2 ** 2 * u2 * v2 * (u2 - v2) ** 2)

    vec = [
        I24 - 2 ** 2 * (u4 * v2 + u2 * v4) * A12 - 2 ** 4 * comb(n - 2, 2) * u6 * A1111,
        I26 - 2 ** 2 * (u6 * v2 + u2 * v6) * A12 - 2 ** 4 * comb(n - 2, 2) * u8 * A1111,
    ]
    A11 = -2 * (n - 2) * A111 + frac(1, 4) * (inv[0][0] * vec[0] + inv[0][1] * vec[1])
    A22 = -2 * (n - 2) * A222 + frac(1, 4) * (inv[1][0] * vec[0] + inv[1][1] * vec[1])

    # mat = [[u2, v2], [u4, v4]]
    det = u2 * v4 - v2 * u4
    inv = [[v4 / det, -v2 / det], [-u4 / det, u2 / det]]
    #
    vec = [
        I2 - 2 ** 4 * comb(n - 1, 3) * u2 * A1111,
        I4 - 2 ** 4 * comb(n - 1, 3) * u4 * A1111,
    ]
    A1 = (
        -2 * (n - 1) * (A11 + A12)
        - 2 ** 2 * comb(n - 1, 2) * A111
        + frac(1, 2) * (inv[0][0] * vec[0] + inv[0][1] * vec[1])
    )
    A2 = (
        -2 * (n - 1) * (A22 + A12)
        - 2 ** 2 * comb(n - 1, 2) * A222
        + frac(1, 2) * (inv[1][0] * vec[0] + inv[1][1] * vec[1])
    )

    A0 = (
        I0
        - 2 * n * (A1 + A2)
        - 2 ** 2 * comb(n, 2) * (A11 + 2 * A12 + A22)
        - 2 ** 3 * comb(n, 3) * (A111 + A222)
        - 2 ** 4 * comb(n, 4) * A1111
    )

    u = sqrt(u2)
    v = sqrt(v2)

    data = [
        (A0, z(n)),
        (A1, fsd(n, (u, 1))),
        (A2, fsd(n, (v, 1))),
        (A11, fsd(n, (u, 2))),
        (A12, fsd(n, (u, 1), (v, 1))),
        (A22, fsd(n, (v, 2))),
        (A111, fsd(n, (u, 3))),
        (A222, fsd(n, (v, 3))),
        (A1111, fsd(n, (u, 4))),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)

    weights /= I0
    variant = "a" if not switch_uv else "b"
    return f"McNamee-Stenger 9{variant}", n, weights, points, 9, _source


# Here starts the cube-specific section.
def integrator(n, k, symbolic):
    """Returns the integral of the polynomial given by the coefficients k over the
    n-dimensional cube.
    """
    assert len(k) <= n

    if any(kk % 2 == 1 for kk in k):
        return 0

    frac = sympy.Rational if symbolic else lambda a, b: a / b
    return frac(2 ** n, prod([kk + 1 for kk in k]))


def mcnamee_stenger_3(n, symbolic=False):
    return CnScheme(*_mcnamee_stenger_3(n, integrator, symbolic=symbolic), 6.374e-14)


def mcnamee_stenger_5(n, symbolic=False):
    return CnScheme(*_mcnamee_stenger_5(n, integrator, symbolic=symbolic), 4.187e-14)


def mcnamee_stenger_7a(n, symbolic=False):
    return CnScheme(
        *_mcnamee_stenger_7(n, integrator, False, symbolic=symbolic), 4.748e-11
    )


def mcnamee_stenger_7b(n, symbolic=False):
    return CnScheme(
        *_mcnamee_stenger_7(n, integrator, True, symbolic=symbolic), 1.669e-12
    )


def mcnamee_stenger_9a(n, symbolic=False):
    return CnScheme(
        *_mcnamee_stenger_9(n, integrator, False, symbolic=symbolic), 3.961e-12
    )


def mcnamee_stenger_9b(n, symbolic=False):
    return CnScheme(
        *_mcnamee_stenger_9(n, integrator, True, symbolic=symbolic), 1.512e-12
    )



import numpy as np

from .. import un
from ..helpers import book
from ._albrecht_collatz import albrecht_collatz_1 as stroud_u3_5_1
from ._albrecht_collatz import albrecht_collatz_2 as stroud_u3_5_2
from ._albrecht_collatz import albrecht_collatz_3 as stroud_u3_5_3
from ._albrecht_collatz import albrecht_collatz_4 as stroud_u3_5_4
from ._albrecht_collatz import albrecht_collatz_5 as stroud_u3_7_2
from ._helpers import U3Scheme, register
from ._mclaren import mclaren_01 as stroud_u3_3_1
from ._mclaren import mclaren_02 as stroud_u3_5_5
from ._mclaren import mclaren_03 as stroud_u3_7_1
from ._mclaren import mclaren_04 as stroud_u3_8_1
from ._mclaren import mclaren_05 as stroud_u3_9_1
from ._mclaren import mclaren_06 as stroud_u3_9_2
from ._mclaren import mclaren_07 as stroud_u3_9_3
from ._mclaren import mclaren_08 as stroud_u3_11_1
from ._mclaren import mclaren_09 as stroud_u3_11_3
from ._mclaren import mclaren_10 as stroud_u3_14_1

source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_u3_11_2():
    scheme = un.stroud_1969(3)
    pts = np.ascontiguousarray(scheme.points.T)
    return U3Scheme(
        "Stroud U3 11-2",
        {"plain": np.vstack([scheme.weights, pts])},
        scheme.degree,
        source,
        4.270e-14,
    )


register(
    [
        stroud_u3_3_1,
        stroud_u3_5_1,
        stroud_u3_5_2,
        stroud_u3_5_3,
        stroud_u3_5_4,
        stroud_u3_5_5,
        stroud_u3_7_2,
        stroud_u3_7_1,
        stroud_u3_8_1,
        stroud_u3_9_1,
        stroud_u3_9_2,
        stroud_u3_9_3,
        stroud_u3_11_1,
        stroud_u3_11_2,
        stroud_u3_11_3,
        stroud_u3_14_1,
    ]
)



# ENH some closed forms
import pathlib

from sympy import Rational as frac
from sympy import sqrt

from ...helpers import article
from .._helpers import C2Scheme, _read, register
from .._tyler import tyler_2 as dunavant_03

source = article(
    authors=["D.A. Dunavant"],
    title="Economical symmetrical quadrature rules for complete polynomials over a square domain",
    journal="Numerical Methods in Engineering",
    volume="21",
    number="10",
    month="oct",
    year="1985",
    pages="1777–1784",
    url="https://doi.org/10.1002/nme.1620211004",
)

this_dir = pathlib.Path(__file__).resolve().parent


def dunavant_00():
    d = {"zero2": [[1]]}
    return C2Scheme("Dunavant 0", d, 1, source, 1.0e-100)


def dunavant_01():
    d = {"d4_aa": [[frac(1, 4)], [sqrt(frac(1, 3))]]}
    return C2Scheme("Dunavant 1", d, 3, source, 4.441e-16)


def dunavant_02():
    d = {
        "d4_a0": [[frac(10, 49)], [sqrt(frac(7, 15))]],
        "d4_aa": [[frac(9, 196)], [sqrt(frac(7, 9))]],
    }
    return C2Scheme("Dunavant 2", d, 5, source, 1.305e-15)


def dunavant_04():
    return _read(this_dir / "dunavant_04.json", source)


def dunavant_05():
    return _read(this_dir / "dunavant_05.json", source)


def dunavant_06():
    return _read(this_dir / "dunavant_06.json", source)


def dunavant_07():
    return _read(this_dir / "dunavant_07.json", source)


def dunavant_08():
    return _read(this_dir / "dunavant_08.json", source)


def dunavant_09():
    return _read(this_dir / "dunavant_09.json", source)


def dunavant_10():
    return _read(this_dir / "dunavant_10.json", source)


register(
    [
        dunavant_00,
        dunavant_01,
        dunavant_02,
        dunavant_03,
        dunavant_04,
        dunavant_05,
        dunavant_06,
        dunavant_07,
        dunavant_08,
        dunavant_09,
        dunavant_10,
    ]
)



import numpy as np
import orthopy

from ..tools import scheme_from_rc
from ._helpers import C1Scheme


def gauss_radau(n, a=0.0, b=0.0):
    assert n >= 2
    degree = 2 * n - 1

    rc = orthopy.c1.jacobi.RecurrenceCoefficients("monic", a, b, symbolic=False)
    _, alpha, beta = np.array([rc[k] for k in range(n)]).T
    points, weights = _radau(alpha, beta, rc.int_1, -1.0)
    return C1Scheme("Gauss-Radau", degree, weights, points)


def _radau(alpha, beta, int_1, xr):
    """From <http://www.scientificpython.net/pyblog/radau-quadrature>:
    Compute the Radau nodes and weights with the preassigned node xr.

    Based on the section 7 of the paper

        Some modified matrix eigenvalue problems,
        Gene Golub,
        SIAM Review Vol 15, No. 2, April 1973, pp.318--334.
    """
    from scipy.linalg import solve_banded

    beta[0] = int_1

    n = len(alpha) - 1
    f = np.zeros(n)
    f[-1] = beta[-1]
    A = np.vstack((np.sqrt(beta), alpha - xr))
    J = np.vstack((A[:, 0:-1], A[0, 1:]))
    delta = solve_banded((1, 1), J, f)
    alphar = alpha.copy()
    alphar[-1] = xr + delta[-1]
    x, w = scheme_from_rc(alphar, beta, int_1, mode="numpy")
    return x, w



from sympy import Rational as frac

from ..helpers import article, expand_symmetries
from ._helpers import CnScheme

_source = article(
    authors=["G.W. Tyler"],
    title="Numerical integration of functions of several variables",
    journal="Canad. J. Math.",
    volume="5",
    year="1953",
    pages="393-412",
    url="https://doi.org/10.4153/CJM-1953-044-1",
)


def tyler(n):
    d = {"0": [[frac(3 - n, 3)]], "a0": [[frac(1, 6)], [1]]}
    points, weights = expand_symmetries(d, n)
    return CnScheme("Tyler", n, weights, points, 3, _source)



import math

import ndim
import numpy as np
import sympy

from .. import un
from ..helpers import article, fsd, gamma_n_2, pm, untangle
from ._helpers import Enr2Scheme

source = article(
    authors=["A.H. Stroud"],
    title="Some Seventh Degree Integration Formulas for Symmetric Regions",
    journal="SIAM J. Numer. Anal.",
    volume="4",
    number="1",
    pages="37–44",
    url="https://doi.org/10.1137/0704004",
)


def _stroud_1967_7_2(n, variant_a, symbolic):
    sqrt = sympy.sqrt if symbolic else math.sqrt

    if variant_a:
        # the points/weights are complex-valued for n >= 9; one could permit that
        assert n in [2, 3, 4, 6, 7]
        p_m = +1
    else:
        assert n in [3, 4]
        p_m = -1

    sqrt38n = sqrt(3 * (8 - n))

    r2 = (3 * (8 - n) - p_m * (n - 2) * sqrt38n) / 2 / (5 - n)
    s2 = (3 * n - p_m * 2 * sqrt38n) / 2 / (3 * n - 8)
    t2 = (6 + p_m * sqrt38n) / 2
    B = (8 - n) / r2 ** 3 / 8
    C = 1 / s2 ** 3 / 2 ** (n + 3)
    D = 1 / t2 ** 3 / 16
    A = 1 - 2 * n * B - 2 ** n * C - 2 * n * (n - 1) * D

    r = sqrt(r2)
    s = sqrt(s2)
    t = sqrt(t2)

    data = [
        (A, [n * [0]]),
        (B, fsd(n, (r, 1))),
        (C, pm(n * [s])),
        (D, fsd(n, (t, 2))),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)

    variant = "a" if variant_a else "b"
    return Enr2Scheme(
        f"Stroud 1967-7 2 (variant {variant})", n, weights, points, 7, source
    )


def stroud_1967_7_2a(n, symbolic=False):
    return _stroud_1967_7_2(n, True, symbolic)


def stroud_1967_7_2b(n, symbolic=False):
    return _stroud_1967_7_2(n, False, symbolic)


def stroud_1967_7_4(n, symbolic=False):
    sqrt = sympy.sqrt if symbolic else math.sqrt

    assert n >= 3

    sqrt2n2 = sqrt(2 * (n + 2))
    r1, r2 = (sqrt((n + 2 - p_m * sqrt2n2) / 2) for p_m in [+1, -1])
    g = gamma_n_2(n, symbolic)
    A1, A2 = ((n + 2 + p_m * sqrt2n2) / 4 / (n + 2) * g for p_m in [+1, -1])

    s = un.stroud_1967(n)

    points = np.concatenate([r1 * s.points, r2 * s.points])
    points = np.ascontiguousarray(points.T)
    weights = np.concatenate([A1 * s.weights, A2 * s.weights])

    weights *= ndim.nsphere.volume(n, symbolic=symbolic) / ndim.enr2.volume(
        n, "physicists", symbolic
    )
    return Enr2Scheme("Stroud 1967-7 4", n, weights, points, 7, source)



# The results have later been reproduced in:
#
# H.T. Rathod, K.V. Nagaraja, B. Venkatesudu,
# Symmetric Gauss Legendre quadrature formulas for composite numerical integration over
# a triangular surface,
# Applied Mathematics and Computation 188 (2007) 865–876
# <https://doi.org/10.1016/j.amc.2006.10.041>.
#
# Reported to Elsevier on June 24, 2019.
#

import numpy as np

from ..c1 import gauss_legendre
from ..helpers import article
from ._helpers import T2Scheme, register

source = article(
    authors=["Frank G. Lether"],
    title="Computation of double integrals over a triangle",
    journal="Journal of Computational and Applied Mathematics",
    volume="2",
    number="3",
    month="sep",
    year="1976",
    pages="219–224",
    url="https://doi.org/10.1016/0771-050X%2876%2990008-5",
)


def lether(n):
    gl = gauss_legendre(n)

    w = np.outer((1 + gl.points) * gl.weights, gl.weights) / 4
    x = np.outer(1 - gl.points, np.ones(n)) / 2
    y = np.outer(1 + gl.points, 1 - gl.points) / 4

    points = np.array([x.flatten(), y.flatten()])
    weights = w.flatten()

    points = np.array([points[0], points[1], 1 - points[0] - points[1]])

    degree = 2 * (n - 1)
    return T2Scheme(
        f"Lether({n})", {"plain": np.vstack([weights, points])}, degree, source
    )


register([lether])



from __future__ import annotations

from math import pi
from typing import Callable

import numpy as np

from .._exception import QuadpyError
from ..helpers import QuadratureScheme, backend_to_function, expand_symmetries

schemes = {}


def register(in_schemes):
    for scheme in in_schemes:
        schemes[scheme.__name__] = scheme


class S3Scheme(QuadratureScheme):
    def __init__(self, name, symmetry_data, degree, source=None, tol=1.0e-14):
        self.symmetry_data = symmetry_data
        points, weights = expand_symmetries(symmetry_data, dim=3)
        super().__init__(name, weights, points, degree, source, tol)
        self.domain = "S3"

    def show(self, backend="vtk", **kwargs):
        """Displays scheme for 3D ball quadrature."""
        return backend_to_function[backend](
            self.points,
            self.weights,
            volume=4.0 / 3.0 * pi,
            edges=[],
            balls=[((0.0, 0.0, 0.0), 1.0)],
            **kwargs,
        )

    def integrate(self, f: Callable, center, radius, dot=np.dot):
        center = np.asarray(center)
        rr = np.multiply.outer(radius, self.points.T)
        rr = np.swapaxes(rr, 0, -2)
        x = (rr + center).T
        ff = np.asarray(f(x))
        if ff.shape[-len(x.shape[1:]) :] != x.shape[1:]:
            string = ", ".join(str(val) for val in x.shape[1:])
            raise QuadpyError(
                f"Wrong return value shape {ff.shape}. Expected (..., {string})."
            )
        return 4 / 3 * pi * np.asarray(radius) ** 3 * dot(ff, self.weights)


def get_good_scheme(degree: int) -> S3Scheme | None:
    if degree <= 7:
        return {
            0: schemes["midpoint"],
            1: schemes["midpoint"],
            2: schemes["hammer_stroud_11_3"],
            3: schemes["hammer_stroud_11_3"],
            4: schemes["hammer_stroud_12_3"],
            5: schemes["hammer_stroud_12_3"],
            6: schemes["mysovskih"],
            7: schemes["mysovskih"],
            8: schemes["stroud_14_1"],
            9: schemes["stroud_14_1"],
            10: schemes["stroud_14_1"],
            11: schemes["stroud_14_1"],
            12: schemes["stroud_14_1"],
            13: schemes["stroud_14_1"],
            14: schemes["stroud_14_1"],
        }[degree]()
    return None



from ._grundmann_moeller import grundmann_moeller
from ._hammer_stroud import hammer_stroud_1a, hammer_stroud_1b, hammer_stroud_2
from ._helpers import TnScheme, get_vol, transform
from ._silvester import silvester
from ._stroud import (
    stroud_tn_1_1,
    stroud_tn_1_2,
    stroud_tn_2_1a,
    stroud_tn_2_1b,
    stroud_tn_2_2,
    stroud_tn_3_1,
    stroud_tn_3_2,
    stroud_tn_3_3,
    stroud_tn_3_4,
    stroud_tn_3_5,
    stroud_tn_3_6a,
    stroud_tn_3_6b,
    stroud_tn_3_7,
    stroud_tn_3_8,
    stroud_tn_3_9,
    stroud_tn_3_10,
    stroud_tn_3_11,
    stroud_tn_4_1,
    stroud_tn_5_1,
    stroud_tn_5_2,
)
from ._stroud_1961 import stroud_1961
from ._stroud_1964 import stroud_1964a, stroud_1964b
from ._stroud_1966 import (
    stroud_1966_1,
    stroud_1966_2,
    stroud_1966_3,
    stroud_1966_4,
    stroud_1966_5,
    stroud_1966_6,
    stroud_1966_7,
)
from ._stroud_1969 import stroud_1969
from ._walkington import (
    walkington_1,
    walkington_2,
    walkington_3,
    walkington_5,
    walkington_7,
)

__all__ = [
    "grundmann_moeller",
    "hammer_stroud_1a",
    "hammer_stroud_1b",
    "hammer_stroud_2",
    "silvester",
    "stroud_1961",
    "stroud_1964a",
    "stroud_1964b",
    "stroud_1966_1",
    "stroud_1966_2",
    "stroud_1966_3",
    "stroud_1966_4",
    "stroud_1966_5",
    "stroud_1966_6",
    "stroud_1966_7",
    "stroud_1969",
    "stroud_tn_1_1",
    "stroud_tn_1_2",
    "stroud_tn_2_1a",
    "stroud_tn_2_1b",
    "stroud_tn_2_2",
    "stroud_tn_3_1",
    "stroud_tn_3_2",
    "stroud_tn_3_3",
    "stroud_tn_3_4",
    "stroud_tn_3_5",
    "stroud_tn_3_6a",
    "stroud_tn_3_6b",
    "stroud_tn_3_7",
    "stroud_tn_3_8",
    "stroud_tn_3_9",
    "stroud_tn_3_10",
    "stroud_tn_3_11",
    "stroud_tn_4_1",
    "stroud_tn_5_1",
    "stroud_tn_5_2",
    "walkington_1",
    "walkington_2",
    "walkington_3",
    "walkington_5",
    "walkington_7",
    #
    "transform",
    "get_vol",
    "TnScheme",
]



import numpy as np
from sympy import Rational as frac
from sympy import cos, pi, sin, sqrt

from ..helpers import article
from ._helpers import Enr2Scheme

source = article(
    authors=["Dongbin Xiu"],
    title="Numerical integration formulas of degree two",
    journal="Applied Numerical Mathematics",
    volume="58",
    year="2008",
    pages="1515–1520",
    url="https://doi.org/10.1016/j.apnum.2007.09.004",
)


def xiu(n):
    points = []
    for k in range(n + 1):
        pt = []
        # Slight adaptation:
        # The article has points for the weight 1/sqrt(2*pi) exp(−x**2/2)
        # so divide by sqrt(2) to adapt for 1/sqrt(pi) exp(−x ** 2)
        for r in range(1, n // 2 + 1):
            alpha = (2 * r * k * pi) / (n + 1)
            pt += [cos(alpha), sin(alpha)]
        if n % 2 == 1:
            pt += [(-1) ** k / sqrt(2)]
        points.append(pt)

    points = np.array(points)
    points = np.ascontiguousarray(points.T)
    weights = np.full(n + 1, frac(1, n + 1))
    return Enr2Scheme("Xiu", n, weights, points, 2, source)



from math import factorial

import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import techreport, untangle
from ._helpers import TnScheme

source = techreport(
    authors=["Noel J. Walkington"],
    title="Quadrature on simplices of arbitrary dimension",
    institution="CMU",
    year="2000",
    url="https://www.math.cmu.edu/~nw0z/publications/00-CNA-023/023abs/",
)


def walkington_1(d):
    degree = 1
    data = [(frac(1, factorial(d)), _c(d, frac))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    # normalize weights
    weights /= np.sum(weights)
    return TnScheme("Walkington 1", d, weights, points, degree, source)


def walkington_2(d):
    # ERR The article claims order 2, but tests really only show order 1.
    # Also, the article says:
    #
    # > The points are inside the simplex when the positive square root is selected.
    #
    # Not sure what this means, but for d>=2, the points are outside the simplex.
    degree = 1
    data = [(frac(1, factorial(d + 1)), _xi1(d, 1 / sqrt(d + 1)))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    # normalize weights
    weights /= np.sum(weights)
    return TnScheme("Walkington 2", d, weights, points, degree, source)


def walkington_3(d):
    degree = 3
    data = [
        (frac(-((d + 1) ** 3), 4 * factorial(d + 2)), _c(d, frac)),
        (frac(+((d + 3) ** 3), 4 * factorial(d + 3)), _xi1(d, frac(1, (d + 3)))),
    ]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    # normalize weights
    weights /= np.sum(weights)
    return TnScheme("Walkington 3", d, weights, points, degree, source)


def walkington_5(d):
    degree = 5
    w0 = frac(+((d + 1) ** 5), 32 * factorial(d + 3))
    w1 = frac(-((d + 3) ** 5), 16 * factorial(d + 4))
    w2 = frac(+((d + 5) ** 5), 16 * factorial(d + 5))
    data = [
        (w0, _c(d, frac)),
        (w1, _xi1(d, frac(1, d + 3))),
        (w2, _xi1(d, frac(1, d + 5))),
        (w2, _xi11(d, frac(1, d + 5), frac)),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    # normalize weights
    weights /= np.sum(weights)
    return TnScheme("Walkington 5", d, weights, points, degree, source)


def walkington_7(d):
    degree = 7
    w0 = -frac(1, 384) * frac((d + 1) ** 7, factorial(d + 4))
    w1 = +frac(1, 128) * frac((d + 3) ** 7, factorial(d + 5))
    w2 = -frac(1, 64) * frac((d + 5) ** 7, factorial(d + 6))
    w3 = +frac(1, 64) * frac((d + 7) ** 7, factorial(d + 7))
    data = [
        (w0, _c(d, frac)),
        (w1, _xi1(d, frac(1, d + 3))),
        (w2, _xi1(d, frac(1, d + 5))),
        (w2, _xi11(d, frac(1, d + 5), frac)),
        (w3, _xi1(d, frac(1, d + 7))),
        (w3, _xi21(d, frac(1, d + 7), frac)),
        (w3, _xi111(d, frac(1, d + 7), frac)),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    # normalize weights
    weights /= np.sum(weights)
    return TnScheme("Walkington 7", d, weights, points, degree, source)


def _c(d, frac):
    return np.array([np.full(d + 1, frac(1, d + 1))])


def _xi1(d, a):
    out = np.full((d + 1, d + 1), a)
    b = 1 - d * a
    np.fill_diagonal(out, b)
    return out


def _xi11(d, a, frac):
    assert d > 1
    b = frac(1 - (d - 1) * a, 2)
    if d == 2:
        out = np.array([[b, b, a], [b, a, b], [a, b, b]])
    else:
        assert d == 3
        out = np.array(
            [
                [b, b, a, a],
                [b, a, b, a],
                [b, a, a, b],
                [a, b, a, b],
                [a, a, b, b],
                [a, b, b, a],
            ]
        )
    return out


def _xi21(d, a, frac):
    assert d > 1
    b = frac(1 - (d - 2) * a, 3)
    # ERR Note that the article incorrectly states (d-2) the the expression for c.
    c = 1 - (d - 1) * a - b
    if d == 2:
        out = np.array(
            [[b, c, a], [c, b, a], [c, a, b], [b, a, c], [a, b, c], [a, c, b]]
        )
    else:
        assert d == 3
        out = np.array(
            [
                [b, c, a, a],
                [b, a, c, a],
                [b, a, a, c],
                [a, b, a, c],
                [a, a, b, c],
                [a, b, c, a],
                [c, b, a, a],
                [c, a, b, a],
                [c, a, a, b],
                [a, c, a, b],
                [a, a, c, b],
                [a, c, b, a],
            ]
        )

    return out


def _xi111(d, a, frac):
    assert d == 3
    b = frac(1 - (d - 2) * a, 3)
    out = np.array([[b, b, b, a], [b, b, a, b], [b, a, b, b], [a, b, b, b]])
    return out



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._classical import centroid as liu_vinokur_01
from ._classical import seven_point as liu_vinokur_07
from ._classical import vertex as liu_vinokur_02
from ._helpers import T2Scheme, register

source = article(
    authors=["Y. Liu", "M. Vinokur"],
    title="Exact Integrations of Polynomials and Symmetric Quadrature Formulas over Arbitrary Polyhedral Grids",
    journal="Journal of Computational Physics",
    volume="140",
    pages="122–147",
    year="1998",
    url="https://doi.org/10.1006/jcph.1998.5884",
)


def liu_vinokur_03():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 2)]]}
    return T2Scheme("Liu-Vinokur 3", d, 2, source)


def liu_vinokur_04():
    d = {"centroid": [[frac(3, 4)]], "vertex": [[frac(1, 12)]]}
    return T2Scheme("Liu-Vinokur 4", d, 2, source)


def liu_vinokur_05():
    # ERR Incorrectly specified in the article as 25 (instead of 2/5).
    # alpha = frac(2, 5)
    # b = (1 - alpha) / 3
    d = {"centroid": [[-frac(9, 16)]], "d3_aa": [[frac(25, 48)], [frac(1, 5)]]}
    return T2Scheme("Liu-Vinokur 5", d, 3, source)


def liu_vinokur_06():
    sqrt21 = sqrt(21)
    alpha1 = (1 - sqrt21) / 10
    b1 = (1 - alpha1) / 3
    d = {"vertex": [[(1 + sqrt21) / 120]], "d3_aa": [[(39 - sqrt21) / 120], [b1]]}
    return T2Scheme("Liu-Vinokur 6", d, 3, source)


def liu_vinokur_08():
    sqrt10 = sqrt(10)
    sqrt_b = sqrt(950 - 220 * sqrt10)
    a1 = (-10 + 5 * sqrt10 + sqrt_b) / 30
    a2 = (-10 + 5 * sqrt10 - sqrt_b) / 30
    b1 = (1 - a1) / 3
    b2 = (1 - a2) / 3
    d = {
        "d3_aa": [
            [
                (5 * a2 - 2) / (60 * a1 ** 2 * (a2 - a1)),
                (5 * a1 - 2) / (60 * a2 ** 2 * (a1 - a2)),
            ],
            [b1, b2],
        ]
    }
    return T2Scheme("Liu-Vinokur 8", d, 4, source)


def liu_vinokur_09():
    alpha0 = -frac(1, 2)
    alpha1 = frac(2, 3)
    b0 = (1 - alpha0) / 3
    b1 = (1 - alpha1) / 3
    d = {
        "centroid": [[frac(27, 80)]],
        "d3_aa": [[frac(8, 105), frac(81, 560)], [b0, b1]],
    }
    return T2Scheme("Liu-Vinokur 9", d, 4, source)


def liu_vinokur_10():
    sqrt13 = sqrt(13)

    alpha0 = 1
    alpha1 = -frac(1, 2)
    alpha2 = (-1 + sqrt13) / 6
    b0 = (1 - alpha0) / 3
    b1 = (1 - alpha1) / 3
    b2 = (1 - alpha2) / 3
    d = {
        "d3_aa": [
            [
                (11 - 1 * sqrt13) / 360,
                (80 - 16 * sqrt13) / 360,
                (29 + 17 * sqrt13) / 360,
            ],
            [b0, b1, b2],
        ]
    }
    return T2Scheme("Liu-Vinokur 10", d, 4, source)


def liu_vinokur_11():
    sqrt3 = sqrt(3)

    gamma = (3 + sqrt3) / 6
    delta = (3 - sqrt3) / 6
    a = (1 + 2 * gamma - delta) / 3
    b = (1 + 2 * delta - gamma) / 3
    # c = (1 - gamma - delta) / 3

    d = {
        "centroid": [[frac(9, 20)]],
        "vertex": [[-frac(1, 60)]],
        "d3_ab": [[frac(1, 10)], [a], [b]],
    }
    return T2Scheme("Liu-Vinokur 11", d, 4, source)


def liu_vinokur_12():
    sqrt15 = sqrt(15)

    a0 = (1 + sqrt15) / 7
    a1 = (1 - sqrt15) / 7
    b0 = (1 - a0) / 3
    b1 = (1 - a1) / 3
    d = {
        "centroid": [[frac(9, 40)]],
        "d3_aa": [[(155 - sqrt15) / 1200, (155 + sqrt15) / 1200], [b0, b1]],
    }
    return T2Scheme("Liu-Vinokur 12", d, 5, source)


def liu_vinokur_13():
    d = {
        "centroid": [[frac(81, 320)]],
        "vertex": [[frac(1, 90)]],
        "d3_aa": [[frac(16, 225), frac(2401, 14400)], [frac(1, 2), frac(1, 7)]],
    }
    return T2Scheme("Liu-Vinokur 13", d, 5, source)


register(
    [
        liu_vinokur_01,
        liu_vinokur_02,
        liu_vinokur_03,
        liu_vinokur_04,
        liu_vinokur_05,
        liu_vinokur_06,
        liu_vinokur_07,
        liu_vinokur_08,
        liu_vinokur_09,
        liu_vinokur_10,
        liu_vinokur_11,
        liu_vinokur_12,
        liu_vinokur_13,
    ]
)



import numpy as np
from sympy import Rational as frac
from sympy import cos, pi, sin, sqrt

from ..helpers import article, untangle
from ._helpers import CnScheme, _s

_source = article(
    authors=["A.H. Stroud"],
    title="Remarks on the Disposition of Points in Numerical Integration Formulas",
    journal="Mathematical Tables and Other Aids to Computation",
    volume="11",
    number="60",
    month="oct",
    year="1957",
    pages="257-261",
    url="https://doi.org/10.2307/2001945",
)


def stroud_1957_2(n):
    r = sqrt(3) / 6
    data = [
        (1.0, np.array([np.full(n, 2 * r)])),
        (+r, _s(n, -1, r)),
        (-r, _s(n, +1, r)),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud 1957-2", n, weights, points, 2, _source, 1.511e-14)


def stroud_1957_3(n):
    n2 = n // 2 if n % 2 == 0 else (n - 1) // 2
    i_range = range(1, 2 * n + 1)
    pts = [
        [
            [sqrt(frac(2, 3)) * cos((2 * k - 1) * i * pi / n) for i in i_range],
            [sqrt(frac(2, 3)) * sin((2 * k - 1) * i * pi / n) for i in i_range],
        ]
        for k in range(1, n2 + 1)
    ]
    if n % 2 == 1:
        sqrt3pm = np.full(2 * n, 1 / sqrt(3))
        sqrt3pm[1::2] *= -1
        pts.append(sqrt3pm)
    pts = np.vstack(pts).T

    data = [(frac(1, 2 * n), pts)]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud 1957-3", n, weights, points, 3, _source)



from ._felippa import felippa_1, felippa_2, felippa_3, felippa_4, felippa_5, felippa_6
from ._kubatko_yeager_maggi import (
    kubatko_yeager_maggi_1,
    kubatko_yeager_maggi_2a,
    kubatko_yeager_maggi_2b,
    kubatko_yeager_maggi_3a,
    kubatko_yeager_maggi_3b,
    kubatko_yeager_maggi_3c,
    kubatko_yeager_maggi_3d,
    kubatko_yeager_maggi_4a,
    kubatko_yeager_maggi_4b,
    kubatko_yeager_maggi_5a,
    kubatko_yeager_maggi_5b,
    kubatko_yeager_maggi_5c,
    kubatko_yeager_maggi_6a,
    kubatko_yeager_maggi_6b,
    kubatko_yeager_maggi_6c,
    kubatko_yeager_maggi_7a,
    kubatko_yeager_maggi_7b,
    kubatko_yeager_maggi_7c,
    kubatko_yeager_maggi_8a,
    kubatko_yeager_maggi_8b,
    kubatko_yeager_maggi_9,
)

__all__ = [
    "felippa_1",
    "felippa_2",
    "felippa_3",
    "felippa_4",
    "felippa_5",
    "felippa_6",
    "kubatko_yeager_maggi_1",
    "kubatko_yeager_maggi_2a",
    "kubatko_yeager_maggi_2b",
    "kubatko_yeager_maggi_3a",
    "kubatko_yeager_maggi_3b",
    "kubatko_yeager_maggi_3c",
    "kubatko_yeager_maggi_3d",
    "kubatko_yeager_maggi_4a",
    "kubatko_yeager_maggi_4b",
    "kubatko_yeager_maggi_5a",
    "kubatko_yeager_maggi_5b",
    "kubatko_yeager_maggi_5c",
    "kubatko_yeager_maggi_6a",
    "kubatko_yeager_maggi_6b",
    "kubatko_yeager_maggi_6c",
    "kubatko_yeager_maggi_7a",
    "kubatko_yeager_maggi_7b",
    "kubatko_yeager_maggi_7c",
    "kubatko_yeager_maggi_8a",
    "kubatko_yeager_maggi_8b",
    "kubatko_yeager_maggi_9",
]



import math

import numpy as np

__all__ = ["plot_disks_1d", "plot_disks"]


def plot_disks_1d(plt, pts, weights, total_area):
    """Plot a circles at quadrature points according to weights. The diameters
    sum up to the total area.
    """
    radii = 0.5 * abs(weights) / math.fsum(weights) * total_area
    colors = ["tab:blue" if weight >= 0 else "tab:red" for weight in weights]
    _plot_disks_helpers(plt, pts, radii, colors)


def plot_disks(plt, pts, weights, total_area):
    """Plot a circles at quadrature points according to weights."""
    flt = np.vectorize(float)
    pts = flt(pts)
    weights = flt(weights)
    radii = np.sqrt(abs(weights) / math.fsum(weights) * total_area / math.pi)
    colors = [
        # use matplotlib 2.0's color scheme
        "tab:blue" if weight >= 0 else "tab:red"
        for weight in weights
    ]
    _plot_disks_helpers(plt, pts, radii, colors)


def _plot_disks_helpers(plt, pts, radii, colors):
    for pt, radius, color in zip(pts, radii, colors):
        # highlight circle center
        plt.plot([pt[0]], [pt[1]], linestyle="None", marker=".", color=color)
        # Choose radius such that the sum of areas of the circles equals total_area.
        # Make sure to set the line width to 0,
        # <https://github.com/matplotlib/matplotlib/issues/17421>.
        circ = plt.Circle((pt[0], pt[1]), radius, color=color, alpha=0.5, linewidth=0)
        plt.gca().add_patch(circ)


def show_mpl(points, weights, volume, edges, balls=None):
    from matplotlib import pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D

    flt = np.vectorize(float)
    points = flt(points)
    weights = flt(weights)

    def plot_spheres(plt, ax, pts, radii, colors):
        h = 1.0e-2

        for tp, r, color in zip(pts, radii, colors):
            # https://matplotlib.org/examples/mplot3d/surface3d_demo2.html
            # Compute sphere for every point anew. This is more costly on the
            # numerical side, but gives the flexibility of drawing sphere of
            # different size with different number of points. Another options
            # would be to precompute x, y, z before the loop, but this can be
            # heavy on the graphics output. See
            # <https://stackoverflow.com/q/45324258/353337>.
            u = np.linspace(0, 2 * np.pi, int(2 * np.pi / h * r) + 1)
            v = np.linspace(0, np.pi, int(np.pi / h * r) + 1)
            _x = np.outer(np.cos(u), np.sin(v))
            _y = np.outer(np.sin(u), np.sin(v))
            _z = np.outer(np.ones(np.size(u)), np.cos(v))

            # highlight ball center
            plt.plot(
                [tp[0]], [tp[1]], [tp[2]], linestyle="None", marker=".", color=color
            )

            ax.plot_surface(
                r * _x + tp[0],
                r * _y + tp[1],
                r * _z + tp[2],
                color=color,
                alpha=0.3,
                linewidth=1,
            )

        ax.set_axis_off()
        return

    balls = [] if balls is None else balls

    ax = plt.axes(projection=Axes3D.name)
    # ax.set_aspect("equal")
    ax.set_axis_off()

    for edge in edges:
        plt.plot(*edge, color="k", linestyle="-")

    plot_spheres(
        plt,
        ax,
        points,
        # Choose radius such that the sum of volumes of the balls equals
        # total_volume.
        radii=np.cbrt(abs(weights) / math.fsum(weights) * volume / (4.0 / 3.0 * np.pi)),
        colors=["tab:blue" if weight >= 0 else "tab:red" for weight in weights],
    )

    for ball in balls:
        plot_spheres(plt, ax, [ball[0]], [ball[1]], ["#dddddd"])

    plt.show()
    return plt


# def show_mayavi(points, weights, volume, edges, balls=None):
#     import mayavi.mlab as mlab
#
#     mlab.figure(bgcolor=(1.0, 1.0, 1.0))
#
#     for edge in edges:
#         mlab.plot3d(*edge, tube_radius=0.5e-2, color=(0.0, 0.0, 0.0))
#
#     blue = (31.0 / 255.0, 119.0 / 255.0, 180.0 / 255.0)
#     red = (84.0 / 255.0, 15.0 / 255.0, 16.0 / 255.0)
#
#     h = 1.0e-2
#     sum_weights = math.fsum(weights)
#     for tp, weight in zip(points, weights):
#         # Choose radius such that the sum of volumes of the balls equals
#         # total_volume.
#         r = (abs(weight) / sum_weights * volume / (4.0 / 3.0 * np.pi)) ** (1.0 / 3.0)
#
#         # Create a sphere
#         u = np.linspace(0, 2 * np.pi, int(2 * np.pi / h * r) + 1)
#         v = np.linspace(0, np.pi, int(np.pi / h * r) + 1)
#         sin_u, cos_u = np.sin(u), np.cos(u)
#         sin_v, cos_v = np.sin(v), np.cos(v)
#         _x = np.outer(cos_u, sin_v)
#         _y = np.outer(sin_u, sin_v)
#         _z = np.outer(np.ones(np.size(u)), cos_v)
#
#         mlab.mesh(
#             r * _x + tp[0],
#             r * _y + tp[1],
#             r * _z + tp[2],
#             color=blue if weight >= 0 else red,
#             opacity=1.0,
#         )
#
#     balls = [] if balls is None else balls
#     for ball in balls:
#         tp = ball[0]
#         r = ball[1]
#
#         # Create a sphere
#         u = np.linspace(0, 2 * np.pi, int(2 * np.pi / h * r) + 1)
#         v = np.linspace(0, np.pi, int(np.pi / h * r) + 1)
#         sin_u, cos_u = np.sin(u), np.cos(u)
#         sin_v, cos_v = np.sin(v), np.cos(v)
#         _x = np.outer(cos_u, sin_v)
#         _y = np.outer(sin_u, sin_v)
#         _z = np.outer(np.ones(np.size(u)), cos_v)
#
#         mlab.mesh(
#             r * _x + tp[0], r * _y + tp[1], r * _z + tp[2], color=[0, 0, 0], opacity=1.0
#         )
#
#     mlab.show()
#     return


def show_vtk(points, weights, volume, edges, balls=None, render=True):
    import vtk

    def get_line_actor(x0, x1):
        source = vtk.vtkLineSource()
        source.SetPoint1(x0)
        source.SetPoint2(x1)
        # mapper
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(source.GetOutputPort())
        # actor
        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        # color actor
        actor.GetProperty().SetColor(0, 0, 0)
        return actor

    def get_sphere_actor(x0, r, color, opacity=1.0):
        # Generate polygon data for a sphere
        sphere = vtk.vtkSphereSource()

        sphere.SetCenter(x0)
        sphere.SetRadius(r)

        sphere.SetPhiResolution(100)
        sphere.SetThetaResolution(100)

        # Create a mapper for the sphere data
        sphere_mapper = vtk.vtkPolyDataMapper()
        # sphere_mapper.SetInput(sphere.GetOutput())
        sphere_mapper.SetInputConnection(sphere.GetOutputPort())

        # Connect the mapper to an actor
        sphere_actor = vtk.vtkActor()
        sphere_actor.SetMapper(sphere_mapper)
        sphere_actor.GetProperty().SetColor(color)
        sphere_actor.GetProperty().SetOpacity(opacity)
        return sphere_actor

    flt = np.vectorize(float)
    points = flt(points)
    weights = flt(weights)

    balls = [] if balls is None else balls

    line_actors = [get_line_actor(edge[:, 0], edge[:, 1]) for edge in edges]

    blue = np.array([31.0, 119.0, 180.0]) / 255.0
    red = np.array([84.0, 15.0, 16.0]) / 255.0

    radii = np.cbrt(abs(weights) / math.fsum(weights) * volume / (4.0 / 3.0 * np.pi))
    sphere_actors = [
        get_sphere_actor(pt, radius, color=blue if weight > 0.0 else red)
        for pt, weight, radius in zip(points.T, weights, radii)
    ]

    sphere_actors.extend(
        [
            get_sphere_actor(
                np.array(ball[0]),
                ball[1],
                color=np.array([0.0, 0.0, 0.0]) / 255.0,
                opacity=0.5,
            )
            for ball in balls
        ]
    )

    # Create a renderer and add the sphere actor to it
    renderer = vtk.vtkRenderer()
    renderer.SetBackground(1.0, 1.0, 1.0)
    # Available in more recent versions of VTK
    # <https://vtk.org/doc/nightly/html/classvtkViewport.html#aed4374e05dbbea1692f7c9c865407664>
    # renderer.SetBackgroundAlpha(1.0)
    for sphere_actor in sphere_actors:
        renderer.AddActor(sphere_actor)
    for line_actor in line_actors:
        renderer.AddActor(line_actor)

    # Create a render window
    render_window = vtk.vtkRenderWindow()
    render_window.AddRenderer(renderer)

    # Create an interactor
    interactor = vtk.vtkRenderWindowInteractor()
    interactor.SetRenderWindow(render_window)

    if render:
        # Initialize the interactor and start the rendering loop
        interactor.Initialize()
        render_window.Render()
        interactor.Start()

    # # Screenshot
    # TODO transparent background
    # w2if = vtk.vtkWindowToImageFilter()
    # w2if.SetInput(render_window)
    # w2if.Update()
    # writer = vtk.vtkPNGWriter()
    # writer.SetFileName('screenshot.png')
    # writer.SetInputConnection(w2if.GetOutputPort())
    # writer.Write()
    return


backend_to_function = {
    # "mayavi": show_mayavi,
    "mpl": show_mpl,
    "vtk": show_vtk,
}



import pathlib

from ...helpers import techreport
from .._helpers import _read, register

source = techreport(
    authors=["R. Cools", "A. Haegemans"],
    title="Construction of fully symmetric cubature formulae of degree 4k-3 for fully symmetric planar regions",
    year="1985",
    institution="Dept. of Computer Science, KU Leuven",
    number="Report TW 71",
    url="https://lirias.kuleuven.be/bitstream/123456789/131870/1/TW71.pdf",
)

this_dir = pathlib.Path(__file__).resolve().parent


def cools_haegemans_1985_9_1():
    return _read(this_dir / "cools_haegemans_1985_09_1.json", source)


def cools_haegemans_1985_13_1():
    return _read(this_dir / "cools_haegemans_1985_13_1.json", source)


def cools_haegemans_1985_13_2():
    return _read(this_dir / "cools_haegemans_1985_13_2.json", source)


def cools_haegemans_1985_13_3():
    return _read(this_dir / "cools_haegemans_1985_13_3.json", source)


def cools_haegemans_1985_17_1():
    return _read(this_dir / "cools_haegemans_1985_17_1.json", source)


def cools_haegemans_1985_17_2():
    return _read(this_dir / "cools_haegemans_1985_17_2.json", source)


register(
    [
        cools_haegemans_1985_9_1,
        cools_haegemans_1985_13_1,
        cools_haegemans_1985_13_2,
        cools_haegemans_1985_13_3,
        cools_haegemans_1985_17_1,
        cools_haegemans_1985_17_2,
    ]
)



from math import factorial as fact

import numpy as np
from sympy import Rational as frac

from ..helpers import article, get_all_exponents, untangle
from ._helpers import TnScheme

source = article(
    authors=["A. Grundmann", "H.M. Möller"],
    title="Invariant integration formulas for the n-simplex by combinatorial methods",
    journal="SIAM J. Numer. Anal.",
    volume="15",
    year="1978",
    pages="282-290",
    url="https://doi.org/10.1137/0715019",
)


def grundmann_moeller(n: int, s: int):
    d = 2 * s + 1

    exponents = get_all_exponents(n + 1, s)

    data = [
        (
            frac(
                (-1) ** i * 2 ** (-2 * s) * (d + n - 2 * i) ** d,
                fact(i) * fact(d + n - i),
            ),
            np.array(
                [
                    [frac(2 * p + 1, d + n - 2 * i) for p in part]
                    for part in exponents[s - i]
                ]
            ),
        )
        for i in range(s + 1)
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    weights /= sum(weights)

    name = f"Grundmann-Möller(dim={n}, {s})"
    return TnScheme(name, n, weights, points, d, source)



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["Bernhard Griener", "Hans Joachim Schmid"],
    title="An interactive tool to visualize common zeros of two-dimensional polynomials",
    journal="Journal of Computational and Applied Mathematics",
    volume="112",
    year="1999",
    pages="83-94",
    url="https://doi.org/10.1016/S0377-0427%2899%2900215-0",
)

# source2 = article(
#     authors="G.G. Rasputin",
#     title="Construction of cubature formulas containing prespecied knots",
#     journal="Metody Vychisl.",
#     volume="13",
#     year="1983",
#     pages="122–129",
#     note="in Russian."
# )

this_dir = pathlib.Path(__file__).resolve().parent


def griener_schmid_1():
    # According to the article, this scheme appeared earlier in `source2`.
    return _read(this_dir / "griener_schmid_1.json", source)


def griener_schmid_2():
    return _read(this_dir / "griener_schmid_2.json", source)


register([griener_schmid_1, griener_schmid_2])



import math

import sympy


def volume(n, lmbda=0, r=1, symbolic=False):
    pi = sympy.pi if symbolic else math.pi
    beta_one_half = (
        (lambda a: sympy.beta(a, sympy.Rational(1, 2)))
        if symbolic
        else (lambda a: math.gamma(a) * math.gamma(0.5) / math.gamma(a + 0.5))
    )

    def _recurrence(n):
        assert n >= 0
        if n == 0:
            return 1
        elif n == 1:
            return beta_one_half(lmbda + 1) * r ** (1 + 2 * lmbda)
        return _recurrence(n - 2) * 2 * pi / (2 * lmbda + n) * r**2

    return _recurrence(n)


def integrate_monomial(exponents, lmbda=0, symbolic=False):
    exponents = list(exponents)
    n = len(exponents)
    n = sympy.S(n) if symbolic else n

    if any(k % 2 == 1 for k in exponents):
        return 0

    def _recurrence(exponents):
        if all(k == 0 for k in exponents):
            return volume(n, lmbda=lmbda, symbolic=symbolic)

        # find first nonzero
        idx, k0 = next((i, k) for i, k in enumerate(exponents) if k > 0)
        k2 = exponents.copy()
        k2[idx] -= 2
        return _recurrence(k2) * (k0 - 1) / (2 * lmbda + sum(exponents) + n)

    return _recurrence(exponents)



from sympy import Rational as frac

from ._helpers import T2Scheme, register


def centroid():
    d = {"centroid": [[1]]}
    return T2Scheme("Centroid rule", d, 1, tol=7.850e-17)


def vertex():
    d = {"vertex": [[frac(1, 3)]]}
    return T2Scheme("Vertex scheme", d, 1)


def seven_point():
    d = {
        "centroid": [[frac(9, 20)]],
        "vertex": [[frac(1, 20)]],
        "d3_aa": [[frac(2, 15)], [frac(1, 2)]],
    }
    return T2Scheme("Seven-point scheme", d, 3)


register([centroid, vertex, seven_point])
